import pytest

import salt.exceptions
import salt.runners.vault as vault
import salt.utils.vault as vaultutil
import salt.utils.vault.api as vapi
import salt.utils.vault.client as vclient
from tests.support.mock import ANY, MagicMock, Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        vault: {
            "__grains__": {"id": "test-master"},
        }
    }


@pytest.fixture
def default_config():
    return {
        "auth": {
            "approle_mount": "approle",
            "approle_name": "salt-master",
            "method": "token",
            "token": "test-token",
            "role_id": "test-role-id",
            "secret_id": None,
            "token_lifecycle": {
                "minimum_ttl": 10,
                "renew_increment": None,
            },
        },
        "cache": {
            "backend": "session",
            "config": 3600,
            "kv_metadata": "connection",
            "secret": "ttl",
        },
        "issue": {
            "allow_minion_override_params": False,
            "type": "token",
            "approle": {
                "mount": "salt-minions",
                "params": {
                    "bind_secret_id": True,
                    "secret_id_num_uses": 1,
                    "secret_id_ttl": 60,
                    "token_explicit_max_ttl": 9999999999,
                    "token_num_uses": 1,
                },
            },
            "token": {
                "role_name": None,
                "params": {
                    "explicit_max_ttl": 9999999999,
                    "num_uses": 1,
                },
            },
            "wrap": "30s",
        },
        "issue_params": {},
        "metadata": {
            "entity": {
                "minion-id": "{minion}",
            },
            "secret": {
                "saltstack-jid": "{jid}",
                "saltstack-minion": "{minion}",
                "saltstack-user": "{user}",
            },
        },
        "policies": {
            "assign": [
                "saltstack/minions",
                "saltstack/{minion}",
            ],
            "cache_time": 60,
            "refresh_pillar": None,
        },
        "server": {
            "url": "http://test-vault:8200",
            "namespace": None,
            "verify": None,
        },
    }


@pytest.fixture
def token_response():
    return {
        "request_id": "0e8c388e-2cb6-bcb2-83b7-625127d568bb",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "auth": {
            "client_token": "test-token",
            "renewable": True,
            "lease_duration": 9999999999,
            "num_uses": 1,
            "creation_time": 1661188581,
            # "expire_time": 11661188580,
        },
    }


@pytest.fixture
def secret_id_response():
    return {
        "request_id": "0e8c388e-2cb6-bcb2-83b7-625127d568bb",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": {
            "secret_id_accessor": "84896a0c-1347-aa90-a4f6-aca8b7558780",
            "secret_id": "841771dc-11c9-bbc7-bcac-6a3945a69cd9",
            "secret_id_ttl": 60,
        },
    }


@pytest.fixture
def wrapped_response():
    return {
        "request_id": "",
        "lease_id": "",
        "lease_duration": 0,
        "renewable": False,
        "data": None,
        "warnings": None,
        "wrap_info": {
            "token": "test-wrapping-token",
            "accessor": "test-wrapping-token-accessor",
            "ttl": 180,
            "creation_time": "2022-09-10T13:37:12.123456789+00:00",
            "creation_path": "whatever/not/checked/here",
            "wrapped_accessor": "84896a0c-1347-aa90-a4f6-aca8b7558780",
        },
    }


@pytest.fixture
def token_serialized(token_response):
    return {
        "client_token": token_response["auth"]["client_token"],
        "renewable": token_response["auth"]["renewable"],
        "lease_duration": token_response["auth"]["lease_duration"],
        "num_uses": token_response["auth"]["num_uses"],
        "creation_time": token_response["auth"]["creation_time"],
        # "expire_time": token_response["auth"]["expire_time"],
    }


@pytest.fixture
def secret_id_serialized(secret_id_response):
    return {
        "secret_id": secret_id_response["data"]["secret_id"],
        "secret_id_ttl": secret_id_response["data"]["secret_id_ttl"],
        "secret_id_num_uses": 1,
        # + creation_time
        # + expire_time
    }


@pytest.fixture
def wrapped_serialized(wrapped_response):
    return {
        "wrap_info": {
            "token": wrapped_response["wrap_info"]["token"],
            "ttl": wrapped_response["wrap_info"]["ttl"],
            "creation_time": 1662817032,
            "creation_path": wrapped_response["wrap_info"]["creation_path"],
        },
    }


@pytest.fixture
def approle_meta(token_serialized, secret_id_serialized):
    return {
        "bind_secret_id": True,
        "local_secret_ids": False,
        "secret_id_bound_cidrs": [],
        "secret_id_num_uses": secret_id_serialized["secret_id_num_uses"],
        "secret_id_ttl": secret_id_serialized["secret_id_ttl"],
        "token_bound_cidrs": [],
        "token_explicit_max_ttl": token_serialized["lease_duration"],
        "token_max_ttl": 0,
        "token_no_default_policy": False,
        "token_num_uses": token_serialized["num_uses"],
        "token_period": 0,
        "token_policies": ["default"],
        "token_ttl": 0,
        "token_type": "default",
    }


@pytest.fixture
def policies_default():
    return ["saltstack/minions", "saltstack/minion/test-minion"]


@pytest.fixture
def metadata_secret_default():
    return {
        "saltstack-jid": "<no jid set>",
        "saltstack-minion": "test-minion",
        "saltstack-user": "<no user set>",
    }


@pytest.fixture
def metadata_entity_default():
    return {"minion-id": "test-minion"}


@pytest.fixture
def grains():
    return {
        "id": "test-minion",
        "roles": ["web", "database"],
        "aux": ["foo", "bar"],
        "deep": {"foo": {"bar": {"baz": ["hello", "world"]}}},
        "mixedcase": "UP-low-UP",
    }


@pytest.fixture
def pillar():
    return {
        "mixedcase": "UP-low-UP",
        "role": "test",
    }


@pytest.fixture
def client():
    with patch("salt.runners.vault._get_master_client", autospec=True) as get_client:
        client = Mock(spec=vclient.AuthenticatedVaultClient)
        get_client.return_value = client
        yield client


@pytest.fixture
def approle_api():
    with patch("salt.runners.vault._get_approle_api", autospec=True) as get_api:
        api = Mock(spec=vapi.AppRoleApi)
        get_api.return_value = api
        yield api


@pytest.fixture
def identity_api():
    with patch("salt.runners.vault._get_identity_api", autospec=True) as get_api:
        api = Mock(spec=vapi.IdentityApi)
        get_api.return_value = api
        yield api


@pytest.fixture
def client_token(client, token_response, wrapped_response):
    def res_or_wrap(*args, **kwargs):
        if kwargs.get("wrap"):
            return vaultutil.VaultWrappedResponse(**wrapped_response["wrap_info"])
        return token_response

    client.post.side_effect = res_or_wrap
    yield client


@pytest.fixture
def config(request, default_config):
    def rec(config, path, val=None, default=vaultutil.VaultException):
        ptr = config
        parts = path.split(":")
        while parts:
            cur = parts.pop(0)
            if val:
                if parts and not isinstance(ptr.get(cur), dict):
                    ptr[cur] = {}
                elif not parts:
                    ptr[cur] = val
                    return
            if cur not in ptr:
                if isinstance(default, Exception):
                    raise default()
                return default
            ptr = ptr[cur]
        return ptr

    def get_config(key=None, default=vaultutil.VaultException):
        overrides = getattr(request, "param", {})
        if key is None:
            for ovar, oval in overrides.items():
                rec(default_config, ovar, oval)
            return default_config
        if key in overrides:
            return overrides[key]
        return rec(default_config, key, default=default)

    with patch("salt.runners.vault._config", autospec=True) as config:
        config.side_effect = get_config
        yield config


@pytest.fixture
def policies(request, policies_default):
    policies_list = getattr(request, "param", policies_default)
    with patch(
        "salt.runners.vault._get_policies_cached", autospec=True
    ) as get_policies_cached:
        get_policies_cached.return_value = policies_list
        with patch("salt.runners.vault._get_policies", autospec=True) as get_policies:
            get_policies.return_value = policies_list
            yield


@pytest.fixture
def metadata(request, metadata_entity_default, metadata_secret_default):
    def _get_metadata(minion_id, metadata_patterns, *args, **kwargs):
        if getattr(request, "param", None) is not None:
            return request.param
        if "saltstack-jid" not in metadata_patterns:
            return metadata_entity_default
        return metadata_secret_default

    with patch("salt.runners.vault._get_metadata", autospec=True) as get_metadata:
        get_metadata.side_effect = _get_metadata
        yield get_metadata


@pytest.fixture
def validate_signature():
    with patch(
        "salt.runners.vault._validate_signature", autospec=True, return_value=None
    ) as validate:
        yield validate


@pytest.mark.usefixtures("policies", "metadata")
@pytest.mark.parametrize(
    "config",
    [{}, {"issue:token:role_name": "test-role"}, {"issue:wrap": False}],
    indirect=True,
)
def test_generate_token(
    client_token,
    config,
    policies_default,
    token_serialized,
    wrapped_serialized,
    metadata_secret_default,
):
    """
    Ensure _generate_token calls the API as expected
    """
    wrap = config("issue:wrap")
    res_token, res_num_uses = vault._generate_token(
        "test-minion", issue_params=None, wrap=wrap
    )
    endpoint = "auth/token/create"
    role_name = config("issue:token:role_name")
    payload = {}
    if config("issue:token:params:explicit_max_ttl"):
        payload["explicit_max_ttl"] = config("issue:token:params:explicit_max_ttl")
    if config("issue:token:params:num_uses"):
        payload["num_uses"] = config("issue:token:params:num_uses")
    payload["meta"] = metadata_secret_default
    payload["policies"] = policies_default
    if role_name:
        endpoint += f"/{role_name}"
    if config("issue:wrap"):
        assert res_token == wrapped_serialized
        client_token.post.assert_called_once_with(
            endpoint, payload=payload, wrap=config("issue:wrap")
        )
    else:
        res_token.pop("expire_time")
        assert res_token == token_serialized
    assert res_num_uses == 1


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize("policies", [[]], indirect=True)
def test_generate_token_no_policies_denied(policies):
    """
    Ensure generated tokens need at least one attached policy
    """
    with pytest.raises(
        salt.exceptions.SaltRunnerError, match=".*No policies matched minion.*"
    ):
        vault._generate_token("test-minion", issue_params=None, wrap=False)


@pytest.mark.parametrize("ttl", [None, 1337])
@pytest.mark.parametrize("uses", [None, 1, 30])
@pytest.mark.parametrize("config", [{}, {"issue:type": "approle"}], indirect=True)
def test_generate_token_deprecated(
    ttl, uses, token_serialized, config, validate_signature, caplog
):
    """
    Ensure the deprecated generate_token function returns data in the old format
    """
    issue_params = {}
    if ttl is not None:
        token_serialized["lease_duration"] = ttl
        issue_params["explicit_max_ttl"] = ttl
    if uses is not None:
        token_serialized["num_uses"] = uses
        issue_params["num_uses"] = uses
    expected = {
        "token": token_serialized["client_token"],
        "lease_duration": token_serialized["lease_duration"],
        "renewable": token_serialized["renewable"],
        "issued": token_serialized["creation_time"],
        "url": config("server:url"),
        "verify": config("server:verify"),
        "token_backend": config("cache:backend"),
        "namespace": config("server:namespace"),
        "uses": token_serialized["num_uses"],
    }
    with patch("salt.runners.vault._generate_token", autospec=True) as gen:
        gen.return_value = (token_serialized, token_serialized["num_uses"])
        res = vault.generate_token("test-minion", "sig", ttl=ttl, uses=uses)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=False
        )
        if config("issue:type") != "token":
            assert "Master is not configured to issue tokens" in caplog.text


@pytest.mark.parametrize("config", [{}, {"issue:wrap": False}], indirect=True)
@pytest.mark.parametrize(
    "issue_params", [None, {"explicit_max_ttl": 120, "num_uses": 3}]
)
def test_generate_new_token(
    issue_params, config, validate_signature, token_serialized, wrapped_serialized
):
    """
    Ensure generate_new_token returns data as expected
    """
    if issue_params is not None:
        if issue_params.get("explicit_max_ttl") is not None:
            token_serialized["lease_duration"] = issue_params["explicit_max_ttl"]
        if issue_params.get("num_uses") is not None:
            token_serialized["num_uses"] = issue_params["num_uses"]
    expected = {"server": config("server"), "auth": {}}
    if config("issue:wrap"):
        expected.update(wrapped_serialized)
        expected.update({"misc_data": {"num_uses": token_serialized["num_uses"]}})
    else:
        expected["auth"] = token_serialized

    with patch("salt.runners.vault._generate_token", autospec=True) as gen:

        def res_or_wrap(*args, **kwargs):
            if kwargs.get("wrap"):
                return wrapped_serialized, token_serialized["num_uses"]
            return token_serialized, token_serialized["num_uses"]

        gen.side_effect = res_or_wrap
        res = vault.generate_new_token("test-minion", "sig", issue_params=issue_params)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=config("issue:wrap")
        )


@pytest.mark.usefixtures("validate_signature")
@pytest.mark.parametrize("config", [{"issue:type": "approle"}], indirect=True)
def test_generate_new_token_refuses_if_not_configured(config):
    """
    Ensure generate_new_token only issues tokens if configured to issue them
    """
    res = vault.generate_new_token("test-minion", "sig")
    assert "error" in res
    assert "Master does not issue tokens" in res["error"]


@pytest.mark.parametrize("config", [{}, {"issue:wrap": False}], indirect=True)
@pytest.mark.parametrize(
    "issue_params", [None, {"explicit_max_ttl": 120, "num_uses": 3}]
)
def test_get_config_token(
    config, validate_signature, token_serialized, wrapped_serialized, issue_params
):
    """
    Ensure get_config returns data in the expected format when configured for token auth
    """
    expected = {
        "auth": {
            "method": "token",
            "token_lifecycle": {
                "minimum_ttl": 10,
                "renew_increment": None,
            },
        },
        "cache": config("cache"),
        "server": config("server"),
        "wrap_info_nested": [],
    }

    if issue_params is not None:
        if issue_params.get("explicit_max_ttl") is not None:
            token_serialized["lease_duration"] = issue_params["explicit_max_ttl"]
        if issue_params.get("num_uses") is not None:
            token_serialized["num_uses"] = issue_params["num_uses"]
    if config("issue:wrap"):
        expected["auth"].update({"token": wrapped_serialized})
        expected.update(
            {
                "wrap_info_nested": ["auth:token"],
                "misc_data": {"token:num_uses": token_serialized["num_uses"]},
            }
        )
    else:
        expected["auth"].update({"token": token_serialized})

    with patch("salt.runners.vault._generate_token", autospec=True) as gen:

        def res_or_wrap(*args, **kwargs):
            if kwargs.get("wrap"):
                return wrapped_serialized, token_serialized["num_uses"]
            return token_serialized, token_serialized["num_uses"]

        gen.side_effect = res_or_wrap
        res = vault.get_config("test-minion", "sig", issue_params=issue_params)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=config("issue:wrap")
        )


@pytest.mark.parametrize(
    "config",
    [
        {"issue:type": "approle"},
        {
            "issue:type": "approle",
            "issue:wrap": False,
            "issue:approle:mount": "test-mount",
        },
        {"issue:type": "approle", "issue:approle:params:bind_secret_id": False},
    ],
    indirect=True,
)
@pytest.mark.parametrize(
    "issue_params",
    [
        None,
        {"token_explicit_max_ttl": 120, "token_num_uses": 3},
        {"secret_id_num_uses": 2, "secret_id_ttl": 120},
    ],
)
def test_get_config_approle(
    config, validate_signature, wrapped_serialized, issue_params
):
    """
    Ensure get_config returns data in the expected format when configured for AppRole auth
    """
    expected = {
        "auth": {
            "approle_mount": config("issue:approle:mount"),
            "approle_name": "test-minion",
            "method": "approle",
            "secret_id": config("issue:approle:params:bind_secret_id"),
            "token_lifecycle": {
                "minimum_ttl": 10,
                "renew_increment": None,
            },
        },
        "cache": config("cache"),
        "server": config("server"),
        "wrap_info_nested": [],
    }

    if config("issue:wrap"):
        expected["auth"].update({"role_id": wrapped_serialized})
        expected.update({"wrap_info_nested": ["auth:role_id"]})
    else:
        expected["auth"].update({"role_id": "test-role-id"})

    with patch("salt.runners.vault._get_role_id", autospec=True) as gen:

        def res_or_wrap(*args, **kwargs):
            if kwargs.get("wrap"):
                return wrapped_serialized
            return "test-role-id"

        gen.side_effect = res_or_wrap
        res = vault.get_config("test-minion", "sig", issue_params=issue_params)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=config("issue:wrap")
        )


@pytest.mark.parametrize(
    "config",
    [{"issue:type": "approle"}, {"issue:type": "approle", "issue:wrap": False}],
    indirect=True,
)
@pytest.mark.parametrize(
    "issue_params",
    [
        None,
        {"token_explicit_max_ttl": 120, "token_num_uses": 3},
        {"secret_id_num_uses": 2, "secret_id_ttl": 120},
    ],
)
def test_get_role_id(config, validate_signature, wrapped_serialized, issue_params):
    """
    Ensure get_role_id returns data in the expected format
    """
    expected = {"server": config("server"), "data": {}}
    if config("issue:wrap"):
        expected.update(wrapped_serialized)
    else:
        expected["data"].update({"role_id": "test-role-id"})
    with patch("salt.runners.vault._get_role_id", autospec=True) as gen:

        def res_or_wrap(*args, **kwargs):
            if kwargs.get("wrap"):
                return wrapped_serialized
            return "test-role-id"

        gen.side_effect = res_or_wrap
        res = vault.get_role_id("test-minion", "sig", issue_params=issue_params)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with(
            "test-minion", issue_params=issue_params or None, wrap=config("issue:wrap")
        )


@pytest.mark.usefixtures("validate_signature")
@pytest.mark.parametrize("config", [{"issue:type": "token"}], indirect=True)
def test_get_role_id_refuses_if_not_configured(config):
    """
    Ensure get_role_id returns an error if not configured to issue AppRoles
    """
    res = vault.get_role_id("test-minion", "sig")
    assert "error" in res
    assert "Master does not issue AppRoles" in res["error"]


class TestGetRoleId:
    @pytest.fixture(autouse=True)
    def lookup_approle(self, approle_meta):
        with patch(
            "salt.runners.vault._lookup_approle_cached", autospec=True
        ) as lookup_approle:
            lookup_approle.return_value = approle_meta
            yield lookup_approle

    @pytest.fixture(autouse=True)
    def lookup_roleid(self, wrapped_serialized):
        role_id = MagicMock(return_value="test-role-id")
        role_id.serialize_for_minion.return_value = wrapped_serialized
        with patch(
            "salt.runners.vault._lookup_role_id", autospec=True
        ) as lookup_roleid:
            lookup_roleid.return_value = role_id
            yield lookup_roleid

    @pytest.fixture(autouse=True)
    def manage_approle(self):
        with patch(
            "salt.runners.vault._manage_approle", autospec=True
        ) as manage_approle:
            yield manage_approle

    @pytest.fixture(autouse=True)
    def manage_entity(self):
        with patch("salt.runners.vault._manage_entity", autospec=True) as manage_entity:
            yield manage_entity

    @pytest.fixture(autouse=True)
    def manage_entity_alias(self):
        with patch(
            "salt.runners.vault._manage_entity_alias", autospec=True
        ) as manage_entity_alias:
            yield manage_entity_alias

    @pytest.mark.parametrize(
        "config",
        [{"issue:type": "approle"}, {"issue:type": "approle", "issue:wrap": False}],
        indirect=True,
    )
    def test_get_role_id(
        self,
        config,
        lookup_approle,
        lookup_roleid,
        manage_approle,
        manage_entity,
        manage_entity_alias,
        wrapped_serialized,
    ):
        """
        Ensure _get_role_id returns data in the expected format and does not
        try to generate a new AppRole if it exists and is configured correctly
        """
        wrap = config("issue:wrap")
        res = vault._get_role_id("test-minion", issue_params=None, wrap=wrap)
        lookup_approle.assert_called_with("test-minion")
        lookup_roleid.assert_called_with("test-minion", wrap=wrap)
        manage_approle.assert_not_called()
        manage_entity.assert_not_called()
        manage_entity_alias.assert_not_called()

        if wrap:
            assert res == wrapped_serialized
            lookup_roleid.return_value.serialize_for_minion.assert_called_once()
        else:
            assert res() == "test-role-id"
            lookup_roleid.return_value.serialize_for_minion.assert_not_called()

    @pytest.mark.parametrize(
        "config",
        [
            {"issue:type": "approle"},
            {"issue:type": "approle", "issue:allow_minion_override_params": True},
        ],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "issue_params", [None, {"token_explicit_max_ttl": 120, "token_num_uses": 3}]
    )
    def test_get_role_id_generate_new(
        self,
        config,
        lookup_approle,
        lookup_roleid,
        manage_approle,
        manage_entity,
        manage_entity_alias,
        wrapped_serialized,
        issue_params,
    ):
        """
        Ensure _get_role_id returns data in the expected format and does not
        try to generate a new AppRole if it exists and is configured correctly
        """
        lookup_approle.return_value = False
        wrap = config("issue:wrap")
        res = vault._get_role_id("test-minion", issue_params=issue_params, wrap=wrap)
        assert res == wrapped_serialized
        lookup_roleid.assert_called_with("test-minion", wrap=wrap)
        manage_approle.assert_called_once_with("test-minion", issue_params)
        manage_entity.assert_called_once_with("test-minion")
        manage_entity_alias.assert_called_once_with("test-minion")

    @pytest.mark.parametrize("config", [{"issue:type": "approle"}], indirect=True)
    def test_get_role_id_generate_new_errors_on_generation_failure(
        self, config, lookup_approle, lookup_roleid
    ):
        """
        Ensure _get_role_id returns an error if the AppRole generation failed
        """
        lookup_approle.return_value = False
        lookup_roleid.return_value = False
        with pytest.raises(
            salt.exceptions.SaltRunnerError,
            match="Failed to create AppRole for minion.*",
        ):
            vault._get_role_id("test-minion", issue_params=None, wrap=False)


@pytest.mark.parametrize(
    "config",
    [{"issue:type": "approle"}, {"issue:type": "approle", "issue:wrap": False}],
    indirect=True,
)
def test_generate_secret_id(
    config, validate_signature, wrapped_serialized, approle_meta, secret_id_serialized
):
    """
    Ensure generate_secret_id returns data in the expected format
    """
    expected = {
        "server": config("server"),
        "data": {},
        "misc_data": {"secret_id_num_uses": approle_meta["secret_id_num_uses"]},
    }
    if config("issue:wrap"):
        expected.update(wrapped_serialized)
    else:
        expected["data"].update(secret_id_serialized)
    with patch("salt.runners.vault._get_secret_id", autospec=True) as gen, patch(
        "salt.runners.vault._approle_params_match", autospec=True, return_value=True
    ) as matcher, patch(
        "salt.runners.vault._lookup_approle_cached", autospec=True
    ) as lookup_approle:

        def res_or_wrap(*args, **kwargs):
            if kwargs.get("wrap"):
                res = Mock(spec=vaultutil.VaultWrappedResponse)
                res.serialize_for_minion.return_value = wrapped_serialized
                return res
            secret_id = Mock(spec=vaultutil.VaultSecretId)
            secret_id.serialize_for_minion.return_value = secret_id_serialized
            return secret_id

        gen.side_effect = res_or_wrap
        lookup_approle.return_value = approle_meta
        res = vault.generate_secret_id("test-minion", "sig", issue_params=None)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with("test-minion", wrap=config("issue:wrap"))
        matcher.assert_called_once()


@pytest.mark.usefixtures("validate_signature")
@pytest.mark.parametrize("config", [{"issue:type": "approle"}], indirect=True)
def test_generate_secret_id_nonexistent_approle(config):
    """
    Ensure generate_secret_id fails and prompts the minion to refresh cache if
    no associated AppRole could be found.
    """
    with patch(
        "salt.runners.vault._lookup_approle_cached", autospec=True
    ) as lookup_approle:
        lookup_approle.return_value = False
        res = vault.generate_secret_id("test-minion", "sig", issue_params=None)
        assert "error" in res
        assert "expire_cache" in res
        assert res["expire_cache"]


@pytest.mark.usefixtures("validate_signature")
@pytest.mark.parametrize("config", [{"issue:type": "token"}], indirect=True)
def test_get_secret_id_refuses_if_not_configured(config):
    """
    Ensure get_secret_id returns an error if not configured to issue AppRoles
    """
    res = vault.generate_secret_id("test-minion", "sig")
    assert "error" in res
    assert "Master does not issue AppRoles" in res["error"]


@pytest.mark.parametrize("config", [{"issue:type": "approle"}], indirect=True)
def test_generate_secret_id_updates_params(
    config, validate_signature, wrapped_serialized, approle_meta
):
    """
    Ensure generate_secret_id returns data in the expected format
    """
    expected = {
        "server": config("server"),
        "data": {},
        "misc_data": {"secret_id_num_uses": approle_meta["secret_id_num_uses"]},
        "wrap_info": wrapped_serialized["wrap_info"],
    }
    with patch("salt.runners.vault._get_secret_id", autospec=True) as gen, patch(
        "salt.runners.vault._approle_params_match", autospec=True, return_value=False
    ) as matcher, patch(
        "salt.runners.vault._manage_approle", autospec=True
    ) as manage_approle, patch(
        "salt.runners.vault._lookup_approle_cached", autospec=True
    ) as lookup_approle:
        res = Mock(spec=vaultutil.VaultWrappedResponse)
        res.serialize_for_minion.return_value = wrapped_serialized
        gen.return_value = res
        lookup_approle.return_value = approle_meta
        res = vault.generate_secret_id("test-minion", "sig", issue_params=None)
        validate_signature.assert_called_once_with("test-minion", "sig", False)
        assert res == expected
        gen.assert_called_once_with("test-minion", wrap=config("issue:wrap"))
        matcher.assert_called_once()
        manage_approle.assert_called_once()


@pytest.mark.parametrize("config", [{"issue:type": "token"}], indirect=True)
def test_list_approles_raises_exception_if_not_configured(config):
    """
    Ensure test_list_approles returns an error if not configured to issue AppRoles
    """
    with pytest.raises(
        salt.exceptions.SaltRunnerError, match="Master does not issue AppRoles.*"
    ):
        vault.list_approles()


@pytest.mark.parametrize(
    "config,expected",
    [
        ({"policies:assign": ["no-tokens-to-replace"]}, ["no-tokens-to-replace"]),
        ({"policies:assign": ["single-dict:{minion}"]}, ["single-dict:test-minion"]),
        (
            {
                "policies:assign": [
                    "should-not-cause-an-exception,but-result-empty:{foo}"
                ]
            },
            [],
        ),
        (
            {"policies:assign": ["Case-Should-Be-Lowered:{grains[mixedcase]}"]},
            ["case-should-be-lowered:up-low-up"],
        ),
        (
            {"policies:assign": ["pillar-rendering:{pillar[role]}"]},
            ["pillar-rendering:test"],
        ),
    ],
    indirect=["config"],
)
def test_get_policies(config, expected, grains, pillar):
    """
    Ensure _get_policies works as intended.
    The expansion of lists is tested in the vault utility module unit tests.
    """
    with patch(
        "salt.utils.minions.get_minion_data",
        MagicMock(return_value=(None, grains, pillar)),
    ):
        with patch(
            "salt.utils.vault.helpers.expand_pattern_lists",
            Mock(side_effect=lambda x, *args, **kwargs: [x]),
        ):
            res = vault._get_policies("test-minion", refresh_pillar=False)
            assert res == expected


@pytest.mark.parametrize(
    "config",
    [
        {"policies:assign": ["salt_minion_{minion}"]},
        {"policies:assign": ["salt_grain_{grains[id]}"]},
        {"policies:assign": ["unset_{foo}"]},
        {"policies:assign": ["salt_pillar_{pillar[role]}"]},
    ],
    indirect=True,
)
def test_get_policies_does_not_render_pillar_unnecessarily(config, grains, pillar):
    """
    The pillar data should only be refreshed in case items are accessed.
    """
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, grains, None)
        with patch(
            "salt.utils.vault.helpers.expand_pattern_lists",
            Mock(side_effect=lambda x, *args, **kwargs: [x]),
        ):
            with patch("salt.pillar.get_pillar", autospec=True) as get_pillar:
                get_pillar.return_value.compile_pillar.return_value = pillar
                vault._get_policies("test-minion", refresh_pillar=True)
                assert get_pillar.call_count == int(
                    "pillar" in config("policies:assign")[0]
                )


@pytest.mark.parametrize(
    "config,expected",
    [
        ({"policies:assign": ["no-tokens-to-replace"]}, ["no-tokens-to-replace"]),
        ({"policies:assign": ["single-dict:{minion}"]}, ["single-dict:test-minion"]),
        ({"policies:assign": ["single-grain:{grains[os]}"]}, []),
    ],
    indirect=["config"],
)
def test_get_policies_for_nonexisting_minions(config, expected):
    """
    For non-existing minions, or the master-minion, grains will be None.
    """
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, None, None)
        with patch(
            "salt.utils.vault.helpers.expand_pattern_lists",
            Mock(side_effect=lambda x, *args, **kwargs: [x]),
        ):
            res = vault._get_policies("test-minion", refresh_pillar=False)
            assert res == expected


@pytest.mark.parametrize(
    "metadata_patterns,expected",
    [
        (
            {"no-tokens-to-replace": "no-tokens-to-replace"},
            {"no-tokens-to-replace": "no-tokens-to-replace"},
        ),
        (
            {"single-dict:{minion}": "single-dict:{minion}"},
            {"single-dict:{minion}": "single-dict:test-minion"},
        ),
        (
            {"should-not-cause-an-exception,but-result-empty:{foo}": "empty:{foo}"},
            {"should-not-cause-an-exception,but-result-empty:{foo}": ""},
        ),
        (
            {
                "Case-Should-Not-Be-Lowered": "Case-Should-Not-Be-Lowered:{pillar[mixedcase]}"
            },
            {"Case-Should-Not-Be-Lowered": "Case-Should-Not-Be-Lowered:UP-low-UP"},
        ),
        (
            {"pillar-rendering:{pillar[role]}": "pillar-rendering:{pillar[role]}"},
            {"pillar-rendering:{pillar[role]}": "pillar-rendering:test"},
        ),
    ],
)
def test_get_metadata(metadata_patterns, expected, pillar):
    """
    Ensure _get_policies works as intended.
    The expansion of lists is tested in the vault utility module unit tests.
    """
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, None, pillar)
        with patch(
            "salt.utils.vault.helpers.expand_pattern_lists",
            Mock(side_effect=lambda x, *args, **kwargs: [x]),
        ):
            res = vault._get_metadata(
                "test-minion", metadata_patterns, refresh_pillar=False
            )
            assert res == expected


def test_get_metadata_list():
    """
    Test that lists are concatenated to an alphabetically sorted
    comma-separated list string since the API does not allow
    composite metadata values
    """
    with patch("salt.utils.minions.get_minion_data", autospec=True) as get_minion_data:
        get_minion_data.return_value = (None, None, None)
        with patch(
            "salt.utils.vault.helpers.expand_pattern_lists", autospec=True
        ) as expand:
            expand.return_value = ["salt_role_foo", "salt_role_bar"]
            res = vault._get_metadata(
                "test-minion",
                {"salt_role": "salt_role_{pillar[roles]}"},
                refresh_pillar=False,
            )
            assert res == {"salt_role": "salt_role_bar,salt_role_foo"}


@pytest.mark.parametrize(
    "config,issue_params,expected",
    [
        (
            {"issue:token:params": {"explicit_max_ttl": None, "num_uses": None}},
            None,
            {},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": 1337, "num_uses": None}},
            None,
            {"explicit_max_ttl": 1337},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": None, "num_uses": 3}},
            None,
            {"num_uses": 3},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": 1337, "num_uses": 3}},
            None,
            {"explicit_max_ttl": 1337, "num_uses": 3},
        ),
        (
            {
                "issue:token:params": {
                    "explicit_max_ttl": 1337,
                    "num_uses": 3,
                    "invalid": True,
                }
            },
            None,
            {"explicit_max_ttl": 1337, "num_uses": 3},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": None, "num_uses": None}},
            {"num_uses": 42, "explicit_max_ttl": 1338},
            {},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": 1337, "num_uses": None}},
            {"num_uses": 42, "explicit_max_ttl": 1338},
            {"explicit_max_ttl": 1337},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": None, "num_uses": 3}},
            {"num_uses": 42, "explicit_max_ttl": 1338},
            {"num_uses": 3},
        ),
        (
            {"issue:token:params": {"explicit_max_ttl": 1337, "num_uses": 3}},
            {"num_uses": 42, "explicit_max_ttl": 1338, "invalid": True},
            {"explicit_max_ttl": 1337, "num_uses": 3},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": None, "num_uses": None},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": None, "explicit_max_ttl": None},
            {},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": None, "num_uses": 3},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": 42, "explicit_max_ttl": None},
            {"num_uses": 42},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": 1337, "num_uses": None},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": None, "explicit_max_ttl": 1338},
            {"explicit_max_ttl": 1338},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": 1337, "num_uses": None},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": 42, "explicit_max_ttl": None},
            {"num_uses": 42, "explicit_max_ttl": 1337},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": None, "num_uses": 3},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": None, "explicit_max_ttl": 1338},
            {"num_uses": 3, "explicit_max_ttl": 1338},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": None, "num_uses": None},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": 42, "explicit_max_ttl": 1338},
            {"num_uses": 42, "explicit_max_ttl": 1338},
        ),
        (
            {
                "issue:token:params": {"explicit_max_ttl": 1337, "num_uses": 3},
                "issue:allow_minion_override_params": True,
            },
            {"num_uses": 42, "explicit_max_ttl": 1338, "invalid": True},
            {"num_uses": 42, "explicit_max_ttl": 1338},
        ),
        ({"issue:type": "approle", "issue:approle:params": {}}, None, {}),
        (
            {
                "issue:type": "approle",
                "issue:approle:params": {
                    "token_explicit_max_ttl": 1337,
                    "token_num_uses": 3,
                    "secret_id_num_uses": 3,
                    "secret_id_ttl": 60,
                },
            },
            None,
            {
                "token_explicit_max_ttl": 1337,
                "token_num_uses": 3,
                "secret_id_num_uses": 3,
                "secret_id_ttl": 60,
            },
        ),
        (
            {
                "issue:type": "approle",
                "issue:approle:params": {
                    "token_explicit_max_ttl": 1337,
                    "token_num_uses": 3,
                    "secret_id_num_uses": 3,
                    "secret_id_ttl": 60,
                },
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
            {
                "token_explicit_max_ttl": 1337,
                "token_num_uses": 3,
                "secret_id_num_uses": 3,
                "secret_id_ttl": 60,
            },
        ),
        (
            {
                "issue:type": "approle",
                "issue:allow_minion_override_params": True,
                "issue:approle:params": {},
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
        ),
        (
            {
                "issue:type": "approle",
                "issue:allow_minion_override_params": True,
                "issue:approle:params": {
                    "token_explicit_max_ttl": 1337,
                    "token_num_uses": 3,
                    "secret_id_num_uses": 3,
                    "secret_id_ttl": 60,
                },
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
            {
                "token_explicit_max_ttl": 1338,
                "token_num_uses": 42,
                "secret_id_num_uses": 42,
                "secret_id_ttl": 1338,
            },
        ),
    ],
    indirect=["config"],
)
def test_parse_issue_params(config, issue_params, expected):
    """
    Ensure all known parameters can only be overridden if it was configured
    on the master. Also ensure the mapping to API requests is correct (for tokens).
    """
    res = vault._parse_issue_params(issue_params)
    assert res == expected


@pytest.mark.parametrize(
    "config,issue_params,expected",
    [
        (
            {"issue:type": "approle", "issue:approle:params": {}},
            {"bind_secret_id": False},
            False,
        ),
        (
            {"issue:type": "approle", "issue:approle:params": {}},
            {"bind_secret_id": True},
            False,
        ),
        (
            {"issue:type": "approle", "issue:approle:params": {"bind_secret_id": True}},
            {"bind_secret_id": False},
            True,
        ),
        (
            {
                "issue:type": "approle",
                "issue:approle:params": {"bind_secret_id": False},
            },
            {"bind_secret_id": True},
            False,
        ),
    ],
    indirect=["config"],
)
def test_parse_issue_params_does_not_allow_bind_secret_id_override(
    config, issue_params, expected
):
    """
    Ensure bind_secret_id can only be set on the master.
    """
    res = vault._parse_issue_params(issue_params)
    assert res.get("bind_secret_id", False) == expected


@pytest.mark.usefixtures("config", "policies")
def test_manage_approle(approle_api, policies_default):
    """
    Ensure _manage_approle calls the API as expected.
    """
    vault._manage_approle("test-minion", None)
    approle_api.write_approle.assert_called_once_with(
        "test-minion",
        mount="salt-minions",
        explicit_max_ttl=9999999999,
        num_uses=1,
        token_policies=policies_default,
    )


@pytest.mark.usefixtures("config")
def test_delete_approle(approle_api):
    """
    Ensure _delete_approle calls the API as expected.
    """
    vault._delete_approle("test-minion")
    approle_api.delete_approle.assert_called_once_with(
        "test-minion", mount="salt-minions"
    )


@pytest.mark.usefixtures("config")
def test_lookup_approle(approle_api, approle_meta):
    """
    Ensure _lookup_approle calls the API as expected.
    """
    approle_api.read_approle.return_value = approle_meta
    res = vault._lookup_approle("test-minion")
    assert res == approle_meta
    approle_api.read_approle.assert_called_once_with(
        "test-minion", mount="salt-minions"
    )


@pytest.mark.usefixtures("config")
def test_lookup_approle_nonexistent(approle_api):
    """
    Ensure _lookup_approle catches VaultNotFoundErrors and returns False.
    """
    approle_api.read_approle.side_effect = vaultutil.VaultNotFoundError
    res = vault._lookup_approle("test-minion")
    assert res is False


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize("wrap", ["30s", False])
def test_lookup_role_id(approle_api, wrap):
    """
    Ensure _lookup_role_id calls the API as expected.
    """
    vault._lookup_role_id("test-minion", wrap=wrap)
    approle_api.read_role_id.assert_called_once_with(
        "test-minion", mount="salt-minions", wrap=wrap
    )


@pytest.mark.usefixtures("config")
def test_lookup_role_id_nonexistent(approle_api):
    """
    Ensure _lookup_role_id catches VaultNotFoundErrors and returns False.
    """
    approle_api.read_role_id.side_effect = vaultutil.VaultNotFoundError
    res = vault._lookup_role_id("test-minion", wrap=False)
    assert res is False


@pytest.mark.usefixtures("config")
@pytest.mark.parametrize("wrap", ["30s", False])
def test_get_secret_id(approle_api, wrap):
    """
    Ensure _get_secret_id calls the API as expected.
    """
    vault._get_secret_id("test-minion", wrap=wrap)
    approle_api.generate_secret_id.assert_called_once_with(
        "test-minion",
        metadata=ANY,
        wrap=wrap,
        mount="salt-minions",
    )


@pytest.mark.usefixtures("config")
def test_lookup_entity_by_alias(identity_api):
    """
    Ensure _lookup_entity_by_alias calls the API as expected.
    """
    with patch("salt.runners.vault._lookup_role_id", return_value="test-role-id"):
        vault._lookup_entity_by_alias("test-minion")
        identity_api.read_entity_by_alias.assert_called_once_with(
            alias="test-role-id", mount="salt-minions"
        )


@pytest.mark.usefixtures("config")
def test_lookup_entity_by_alias_failed(identity_api):
    """
    Ensure _lookup_entity_by_alias returns False if the lookup fails.
    """
    with patch("salt.runners.vault._lookup_role_id", return_value="test-role-id"):
        identity_api.read_entity_by_alias.side_effect = vaultutil.VaultNotFoundError
        res = vault._lookup_entity_by_alias("test-minion")
        assert res is False


@pytest.mark.usefixtures("config")
def test_fetch_entity_by_name(identity_api):
    """
    Ensure _fetch_entity_by_name calls the API as expected.
    """
    vault._fetch_entity_by_name("test-minion")
    identity_api.read_entity.assert_called_once_with(name="salt_minion_test-minion")


@pytest.mark.usefixtures("config")
def test_fetch_entity_by_name_failed(identity_api):
    """
    Ensure _fetch_entity_by_name returns False if the lookup fails.
    """
    identity_api.read_entity.side_effect = vaultutil.VaultNotFoundError
    res = vault._fetch_entity_by_name("test-minion")
    assert res is False


@pytest.mark.usefixtures("config")
def test_manage_entity(identity_api, metadata, metadata_entity_default):
    """
    Ensure _manage_entity calls the API as expected.
    """
    vault._manage_entity("test-minion")
    identity_api.write_entity.assert_called_with(
        "salt_minion_test-minion", metadata=metadata_entity_default
    )


@pytest.mark.usefixtures("config")
def test_delete_entity(identity_api):
    """
    Ensure _delete_entity calls the API as expected.
    """
    vault._delete_entity("test-minion")
    identity_api.delete_entity.assert_called_with("salt_minion_test-minion")


@pytest.mark.usefixtures("config")
def test_manage_entity_alias(identity_api):
    """
    Ensure _manage_entity_alias calls the API as expected.
    """
    with patch("salt.runners.vault._lookup_role_id", return_value="test-role-id"):
        vault._manage_entity_alias("test-minion")
        identity_api.write_entity_alias.assert_called_with(
            "salt_minion_test-minion", alias_name="test-role-id", mount="salt-minions"
        )


@pytest.mark.usefixtures("config")
def test_manage_entity_alias_raises_errors(identity_api):
    """
    Ensure _manage_entity_alias raises exceptions.
    """
    identity_api.write_entity_alias.side_effect = vaultutil.VaultNotFoundError
    with patch("salt.runners.vault._lookup_role_id", return_value="test-role-id"):
        with pytest.raises(
            salt.exceptions.SaltRunnerError,
            match="Cannot create alias.* no entity found.",
        ):
            vault._manage_entity_alias("test-minion")


def test_revoke_token_by_token(client):
    """
    Ensure _revoke_token calls the API as expected.
    """
    vault._revoke_token(token="test-token")
    client.post.assert_called_once_with(
        "auth/token/revoke", payload={"token": "test-token"}
    )


def test_revoke_token_by_accessor(client):
    """
    Ensure _revoke_token calls the API as expected.
    """
    vault._revoke_token(accessor="test-accessor")
    client.post.assert_called_once_with(
        "auth/token/revoke-accessor", payload={"accessor": "test-accessor"}
    )



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.mark.parametrize("test", (True, False))
def test_touch(file, tmp_path, test):
    """
    file.touch
    """
    name = tmp_path / "testfile"
    ret = file.touch(name=str(name), test=test)
    if test is True:
        assert ret.result is None
        assert name.is_file() is False
    else:
        assert ret.result is True
        assert name.is_file()


def test_touch_directory(file, tmp_path):
    """
    file.touch a directory
    """
    name = tmp_path / "touch_test_dir"
    name.mkdir()
    ret = file.touch(name=str(name))
    assert ret.result is True
    assert name.is_dir()



import pytest

from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.destructive_test,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def chocolatey(states):
    yield states.chocolatey


@pytest.fixture(scope="module")
def chocolatey_mod(modules):
    yield modules.chocolatey


@pytest.fixture()
def clean(chocolatey_mod):
    try:
        # If chocolatey is not installed, this will throw an error
        chocolatey_mod.chocolatey_version(refresh=True)
        # If we get this far, chocolatey is installed... let's uninstall
        chocolatey_mod.unbootstrap()
    except CommandExecutionError:
        pass

    # Try to get the new version, should throw an error
    try:
        chocolatey_version = chocolatey_mod.chocolatey_version(refresh=True)
    except CommandExecutionError:
        chocolatey_version = None

    # Assert the chocolatey is not installed
    assert chocolatey_version is None
    try:
        yield
    finally:
        try:
            # If chocolatey is not installed, this will throw an error
            chocolatey_mod.chocolatey_version(refresh=True)
            # If we get this far, chocolatey is installed... let's uninstall
            chocolatey_mod.unbootstrap()
        except CommandExecutionError:
            pass


@pytest.fixture()
def installed(chocolatey_mod):
    # Even though this variable is not used, it will be displayed in test output
    # if there is an error
    result = chocolatey_mod.bootstrap(force=True)
    try:
        chocolatey_version = chocolatey_mod.chocolatey_version(refresh=True)
    except CommandExecutionError:
        chocolatey_version = None
    assert chocolatey_version is not None
    try:
        yield
    finally:
        try:
            # If chocolatey is not installed, this will throw an error
            chocolatey_mod.chocolatey_version(refresh=True)
            # If we get this far, chocolatey is installed... let's uninstall
            chocolatey_mod.unbootstrap()
        except CommandExecutionError:
            pass


def test_bootstrapped(chocolatey, chocolatey_mod, clean):
    ret = chocolatey.bootstrapped(name="junk name")
    assert "Installed chocolatey" in ret.comment
    assert ret.result is True
    try:
        chocolatey_version = chocolatey_mod.chocolatey_version(refresh=True)
    except CommandExecutionError:
        chocolatey_version = None
    assert chocolatey_version is not None


def test_bootstrapped_test_true(chocolatey, clean):
    ret = chocolatey.bootstrapped(name="junk name", test=True)
    assert ret.result is None
    assert ret.comment == "The latest version of Chocolatey will be installed"


def test_bootstrapped_version(chocolatey, chocolatey_mod, clean):
    ret = chocolatey.bootstrapped(name="junk_name", version="1.4.0")
    assert ret.comment == "Installed chocolatey 1.4.0"
    assert ret.result is True
    try:
        chocolatey_version = chocolatey_mod.chocolatey_version(refresh=True)
    except CommandExecutionError:
        chocolatey_version = None
    assert chocolatey_version == "1.4.0"


def test_bootstrapped_version_test_true(chocolatey, chocolatey_mod, clean):
    ret = chocolatey.bootstrapped(name="junk_name", version="1.4.0", test=True)
    assert ret.comment == "Chocolatey 1.4.0 will be installed"


def test_unbootstrapped_installed(chocolatey, chocolatey_mod, installed):
    ret = chocolatey.unbootstrapped(name="junk_name")
    assert "Uninstalled chocolatey" in ret.comment
    assert ret.result is True
    try:
        chocolatey_version = chocolatey_mod.chocolatey_version(refresh=True)
    except CommandExecutionError:
        chocolatey_version = None
    assert chocolatey_version is None


def test_unbootstrapped_installed_test_true(chocolatey, chocolatey_mod, installed):
    ret = chocolatey.unbootstrapped(name="junk_name", test=True)
    assert "will be removed" in ret.comment
    assert ret.result is None


def test_unbootstrapped_clean(chocolatey, chocolatey_mod, clean):
    ret = chocolatey.unbootstrapped(name="junk_name")
    assert ret.comment == "Chocolatey not found on this system"
    assert ret.result is True



"""
Retrieve Pillar data by doing a MySQL query

MariaDB provides Python support through the MySQL Python package.
Therefore, you may use this module with both MySQL or MariaDB.

This module is a concrete implementation of the sql_base ext_pillar for MySQL.

:maturity: new
:depends: python-mysqldb
:platform: all

Configuring the mysql ext_pillar
================================

Use the 'mysql' key under ext_pillar for configuration of queries.

MySQL configuration of the MySQL returner is being used (mysql.db, mysql.user,
mysql.pass, mysql.port, mysql.host) for database connection info.

Required python modules: MySQLdb

Complete example
================

.. code-block:: yaml

    mysql:
      user: 'salt'
      pass: 'super_secret_password'
      db: 'salt_db'
      port: 3306
      ssl:
        cert: /etc/mysql/client-cert.pem
        key: /etc/mysql/client-key.pem

    ext_pillar:
      - mysql:
          fromdb:
            query: 'SELECT col1,col2,col3,col4,col5,col6,col7
                      FROM some_random_table
                     WHERE minion_pattern LIKE %s'
            depth: 5
            as_list: True
            with_lists: [1,3]
"""

import logging
from contextlib import contextmanager

from salt.pillar.sql_base import SqlBaseExtPillar

# Set up logging
log = logging.getLogger(__name__)

try:
    # Trying to import MySQLdb
    import MySQLdb
    import MySQLdb.converters
    import MySQLdb.cursors
except ImportError:
    try:
        # MySQLdb import failed, try to import PyMySQL
        import pymysql

        pymysql.install_as_MySQLdb()
        import MySQLdb
        import MySQLdb.converters
        import MySQLdb.cursors
    except ImportError:
        MySQLdb = None


def __virtual__():
    """
    Confirm that a python mysql client is installed.
    """
    return bool(MySQLdb), "No python mysql client installed." if MySQLdb is None else ""


class MySQLExtPillar(SqlBaseExtPillar):
    """
    This class receives and processes the database rows from MySQL.
    """

    @classmethod
    def _db_name(cls):
        return "MySQL"

    def _get_options(self):
        """
        Returns options used for the MySQL connection.
        """
        defaults = {
            "host": "localhost",
            "user": "salt",
            "pass": "salt",
            "db": "salt",
            "port": 3306,
            "ssl": {},
        }
        _options = {}
        _opts = __opts__.get("mysql", {})
        for attr in defaults:
            if attr not in _opts:
                log.debug("Using default for MySQL %s", attr)
                _options[attr] = defaults[attr]
                continue
            _options[attr] = _opts[attr]
        return _options

    @contextmanager
    def _get_cursor(self):
        """
        Yield a MySQL cursor
        """
        _options = self._get_options()
        conn = MySQLdb.connect(
            host=_options["host"],
            user=_options["user"],
            passwd=_options["pass"],
            db=_options["db"],
            port=_options["port"],
            ssl=_options["ssl"],
        )
        cursor = conn.cursor()
        try:
            yield cursor
        except MySQLdb.DatabaseError as err:
            log.exception("Error in ext_pillar MySQL: %s", err.args)
        finally:
            conn.close()

    def extract_queries(self, args, kwargs):  # pylint: disable=useless-super-delegation
        """
        This function normalizes the config block into a set of queries we
        can use.  The return is a list of consistently laid out dicts.
        """
        return super().extract_queries(args, kwargs)


def ext_pillar(minion_id, pillar, *args, **kwargs):
    """
    Execute queries against MySQL, merge and return as a dict
    """
    return MySQLExtPillar().fetch(minion_id, pillar, *args, **kwargs)



import pytest

from tests.support.pytest.helpers import reap_stray_processes


@pytest.fixture(scope="package", autouse=True)
def _auto_skip_on_system_python_too_recent(grains):
    if (
        grains["osfinger"] in ("Fedora Linux-40", "Ubuntu-24.04")
        or grains["os_family"] == "Arch"
    ):
        pytest.skip(
            "System ships with a version of python that is too recent for salt-ssh tests",
            # Actually, the problem is that the tornado we ship is not prepared for Python 3.12,
            # and it imports `ssl` and checks if the `match_hostname` function is defined, which
            # has been deprecated since Python 3.7, so, the logic goes into trying to import
            # backports.ssl-match-hostname which is not installed on the system.
        )


@pytest.fixture(autouse=True)
def _reap_stray_processes():
    # when tests timeout, we migth leave child processes behind
    # nuke them
    with reap_stray_processes():
        # Run test
        yield


@pytest.fixture(scope="module")
def state_tree(base_env_state_tree_root_dir):
    top_file = """
    {%- from "map.jinja" import abc with context %}
    base:
      'localhost':
        - basic
      '127.0.0.1':
        - basic
    """
    map_file = """
    {%- set abc = "def" %}
    """
    state_file = """
    {%- from "map.jinja" import abc with context %}
    Ok with {{ abc }}:
      test.succeed_with_changes
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file(
        "map.jinja", map_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "test.sls", state_file, base_env_state_tree_root_dir
    )
    with top_tempfile, map_tempfile, state_tempfile:
        yield


@pytest.fixture(scope="module")
def state_tree_dir(base_env_state_tree_root_dir):
    """
    State tree with files to test salt-ssh
    when the map.jinja file is in another directory
    """
    top_file = """
    {%- from "test/map.jinja" import abc with context %}
    base:
      'localhost':
        - test
      '127.0.0.1':
        - test
    """
    map_file = """
    {%- set abc = "def" %}
    """
    state_file = """
    {%- from "test/map.jinja" import abc with context %}

    Ok with {{ abc }}:
      test.succeed_without_changes
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_state_tree_root_dir
    )
    map_tempfile = pytest.helpers.temp_file(
        "test/map.jinja", map_file, base_env_state_tree_root_dir
    )
    state_tempfile = pytest.helpers.temp_file(
        "test.sls", state_file, base_env_state_tree_root_dir
    )

    with top_tempfile, map_tempfile, state_tempfile:
        yield



"""
    :codeauthor: asomers <asomers@gmail.com>
"""

import pytest

import salt.modules.freebsd_sysctl as freebsd_sysctl
import salt.modules.systemd_service as systemd
from salt.exceptions import CommandExecutionError
from tests.support.helpers import dedent
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {freebsd_sysctl: {}, systemd: {}}


def test_get():
    """
    Tests the return of get function
    """
    mock_cmd = MagicMock(return_value="1")
    with patch.dict(freebsd_sysctl.__salt__, {"cmd.run": mock_cmd}):
        assert freebsd_sysctl.get("vfs.usermount") == "1"


def test_assign_failed():
    """
    Tests if the assignment was successful or not
    """
    cmd = {
        "pid": 1337,
        "retcode": 1,
        "stderr": "sysctl: unknown oid 'asef.esrhaseras.easr'",
        "stdout": "",
    }
    mock_cmd = MagicMock(return_value=cmd)
    with patch.dict(freebsd_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
        pytest.raises(
            CommandExecutionError,
            freebsd_sysctl.assign,
            "asef.esrhaseras.easr",
            "backward",
        )


def test_assign_success():
    """
    Tests the return of successful assign function
    """
    cmd = {
        "pid": 1337,
        "retcode": 0,
        "stderr": "",
        "stdout": "vfs.usermount: 0 -> 1",
    }
    ret = {"vfs.usermount": "1"}
    mock_cmd = MagicMock(return_value=cmd)
    with patch.dict(freebsd_sysctl.__salt__, {"cmd.run_all": mock_cmd}):
        assert freebsd_sysctl.assign("vfs.usermount", 1) == ret


def test_persist_no_conf_failure():
    """
    Tests adding of config file failure
    """
    asn_cmd = {
        "pid": 1337,
        "retcode": 1,
        "stderr": "sysctl: vfs.usermount=1: Operation not permitted",
        "stdout": "vfs.usermount: 1",
    }
    mock_asn_cmd = MagicMock(return_value=asn_cmd)
    cmd = "sysctl vfs.usermount=1"
    mock_cmd = MagicMock(return_value=cmd)
    with patch.dict(
        freebsd_sysctl.__salt__,
        {"cmd.run_stdout": mock_cmd, "cmd.run_all": mock_asn_cmd},
    ):
        with patch("salt.utils.files.fopen", mock_open()) as m_open:
            pytest.raises(
                CommandExecutionError,
                freebsd_sysctl.persist,
                "net.ipv4.ip_forward",
                1,
                config=None,
            )


def test_persist_nochange():
    """
    Tests success when no changes need to be made
    """
    mock_get_cmd = MagicMock(return_value="1")
    content = "vfs.usermount=1\n"
    with patch("salt.utils.files.fopen", mock_open(read_data=content)):
        with patch.dict(
            freebsd_sysctl.__salt__,
            {"cmd.run": mock_get_cmd},
        ):
            assert freebsd_sysctl.persist("vfs.usermount", 1) == "Already set"


def test_persist_in_memory():
    """
    Tests success when the on-disk value is correct but the in-memory value
    needs updating.
    """
    mock_get_cmd = MagicMock(return_value="0")
    set_cmd = {
        "pid": 1337,
        "retcode": 0,
        "stderr": "",
        "stdout": "vfs.usermount: 0 -> 1",
    }
    mock_set_cmd = MagicMock(return_value=set_cmd)
    content = "vfs.usermount=1\n"
    with patch("salt.utils.files.fopen", mock_open(read_data=content)):
        with patch.dict(
            freebsd_sysctl.__salt__,
            {"cmd.run": mock_get_cmd},
        ):
            with patch.dict(
                freebsd_sysctl.__salt__,
                {"cmd.run_all": mock_set_cmd},
            ):
                assert freebsd_sysctl.persist("vfs.usermount", 1) == "Updated"


def test_persist_updated():
    """
    Tests sysctl.conf success
    """
    cmd = {
        "pid": 1337,
        "retcode": 0,
        "stderr": "",
        "stdout": "vfs.usermount: 1 -> 1",
    }
    mock_cmd = MagicMock(return_value=cmd)

    with patch("salt.utils.files.fopen", mock_open()):
        with patch.dict(
            freebsd_sysctl.__salt__,
            {"cmd.run_all": mock_cmd},
        ):
            assert freebsd_sysctl.persist("vfs.usermount", 1) == "Updated"


def test_persist_updated_tunable():
    """
    Tests loader.conf success
    """

    with patch("salt.utils.files.fopen", mock_open()):
        assert (
            freebsd_sysctl.persist("vfs.usermount", 1, "/boot/loader.conf") == "Updated"
        )


def test_show():
    """
    Tests the show function
    """
    # Mock just a small portion of the full "sysctl -ae" output, but be
    # sure to include a multi-line value.
    mock_cmd = MagicMock(
        return_value=dedent(
            """\
        kern.ostype=FreeBSD
        kern.osrelease=13.0-CURRENT
        kern.osrevision=199506
        kern.version=FreeBSD 13.0-CURRENT #246 r365916M: Thu Sep 24 09:17:12 MDT 2020
            user@host.domain:/usr/obj/usr/src/head
        /amd64.amd64/sys/GENERIC

        kern.maxvnodes=213989
        """,
            "\n",
        )
    )
    with patch.dict(freebsd_sysctl.__salt__, {"cmd.run": mock_cmd}):
        ret = freebsd_sysctl.show()
        assert "FreeBSD" == ret["kern.ostype"]
        assert (
            dedent(
                """\
            FreeBSD 13.0-CURRENT #246 r365916M: Thu Sep 24 09:17:12 MDT 2020
                user@host.domain:/usr/obj/usr/src/head
            /amd64.amd64/sys/GENERIC
            """,
                "\n",
            )
            == ret["kern.version"]
        )



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
    :codeauthor: Gareth J. Greenaway <greenaway@vmware.com>
"""

import pytest

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.timeout(120, func_only=True),
    pytest.mark.destructive_test,
    pytest.mark.skip_if_not_root,
    pytest.mark.skip_unless_on_darwin,
    pytest.mark.skip_if_binaries_missing("brew"),
]


@pytest.fixture(scope="module")
def pkg(modules):
    return modules.pkg


@pytest.fixture
def pkg_1_name(pkg):
    pkg_name = "algol68g"
    try:
        yield pkg_name
    finally:
        pkg_list = pkg.list_pkgs()

        # Remove package if installed
        if pkg_name in pkg_list:
            pkg.remove(pkg_name)


@pytest.fixture
def pkg_2_name(pkg):
    pkg_name = "acme"
    try:
        pkg.install(pkg_name)
        pkg_list = pkg.list_pkgs()
        if pkg_name not in pkg_list:
            pytest.skip(f"Failed to install the '{pkg_name}' package to delete")
        yield pkg_name
    finally:
        pkg_list = pkg.list_pkgs()

        # Remove package if still installed
        if pkg_name in pkg_list:
            pkg.remove(pkg_name)


def test_brew_install(pkg, pkg_1_name):
    """
    Tests the installation of packages
    """
    pkg.install(pkg_1_name)
    pkg_list = pkg.list_pkgs()
    assert pkg_1_name in pkg_list


def test_remove(pkg, pkg_2_name):
    """
    Tests the removal of packages
    """
    pkg.remove(pkg_2_name)
    pkg_list = pkg.list_pkgs()
    assert pkg_2_name not in pkg_list


def test_version(pkg, pkg_1_name):
    """
    Test pkg.version for mac. Installs a package and then checks we can get
    a version for the installed package.
    """
    pkg.install(pkg_1_name)
    pkg_list = pkg.list_pkgs()
    version = pkg.version(pkg_1_name)
    assert version
    assert pkg_1_name in pkg_list
    # make sure the version is accurate and is listed in the pkg_list
    assert version in str(pkg_list[pkg_1_name])


def test_latest_version(pkg, pkg_1_name):
    """
    Test pkg.latest_version:
      - get the latest version available
      - install the package
      - get the latest version available
      - check that the latest version is empty after installing it
    """
    pkg.remove(pkg_1_name)
    uninstalled_latest = pkg.latest_version(pkg_1_name)

    pkg.install(pkg_1_name)
    installed_latest = pkg.latest_version(pkg_1_name)
    version = pkg.version(pkg_1_name)
    assert isinstance(uninstalled_latest, str)
    assert installed_latest == version


def test_refresh_db(pkg):
    """
    Integration test to ensure pkg.refresh_db works with brew
    """
    refresh_brew = pkg.refresh_db()
    assert refresh_brew


def test_list_upgrades(pkg, pkg_1_name):
    """
    Test pkg.list_upgrades: data is in the form {'name1': 'version1', 'name2': 'version2', ... }
    """
    upgrades = pkg.list_upgrades()
    assert isinstance(upgrades, dict)
    if upgrades:
        for name in upgrades:
            assert isinstance(name, str)
            assert isinstance(upgrades[name], str)


def test_info_installed(pkg, pkg_1_name):
    """
    Test pkg.info_installed: info returned has certain fields used by
    mac_brew.latest_version
    """
    pkg.install(pkg_1_name)
    info = pkg.info_installed(pkg_1_name)
    assert pkg_1_name in info
    assert "versions" in info[pkg_1_name]
    assert "revision" in info[pkg_1_name]
    assert "stable" in info[pkg_1_name]["versions"]



"""
Salt dunders.
"""

import salt.loader.context

loader_context = salt.loader.context.LoaderContext()


__file_client__ = loader_context.named_context("__file_client__", default=None)
__opts__ = loader_context.named_context("__opts__")
__context__ = loader_context.named_context("__context__")
__pillar__ = loader_context.named_context("__pillar__")
__grains__ = loader_context.named_context("__grains__")



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import textwrap
import uuid

import pytest

import salt.modules.iptables as iptables
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {iptables: {}}


# 'version' function tests: 1


def test_version():
    """
    Test if it return version from iptables --version
    """
    mock = MagicMock(return_value="iptables v1.4.21")
    with patch.dict(iptables.__salt__, {"cmd.run_stdout": mock}):
        assert iptables.version() == "v1.4.21"


# 'build_rule' function tests: 1


def test_build_rule():
    """
    Test if it build a well-formatted iptables rule based on kwargs.
    """
    with patch.object(iptables, "_has_option", MagicMock(return_value=True)):
        assert iptables.build_rule() == ""

        assert (
            iptables.build_rule(name="ignored", state="ignored") == ""
        ), "build_rule should ignore name and state"

        # Should properly negate bang-prefixed values
        assert iptables.build_rule(**{"if": "!eth0"}) == "! -i eth0"

        # Should properly negate "not"-prefixed values
        assert iptables.build_rule(**{"if": "not eth0"}) == "! -i eth0"

        assert (
            iptables.build_rule(**{"protocol": "tcp", "syn": "!"}) == "-p tcp ! --syn"
        )

        assert (
            iptables.build_rule(dports=[80, 443], protocol="tcp")
            == "-p tcp -m multiport --dports 80,443"
        )

        assert (
            iptables.build_rule(dports="80,443", protocol="tcp")
            == "-p tcp -m multiport --dports 80,443"
        )

        # Should it really behave this way?
        assert (
            iptables.build_rule(dports=["!80", 443], protocol="tcp")
            == "-p tcp -m multiport ! --dports 80,443"
        )

        assert (
            iptables.build_rule(dports="!80,443", protocol="tcp")
            == "-p tcp -m multiport ! --dports 80,443"
        )

        assert (
            iptables.build_rule(sports=[80, 443], protocol="tcp")
            == "-p tcp -m multiport --sports 80,443"
        )

        assert (
            iptables.build_rule(sports="80,443", protocol="tcp")
            == "-p tcp -m multiport --sports 80,443"
        )

        assert (
            iptables.build_rule(
                "filter",
                "INPUT",
                command="I",
                position="3",
                full=True,
                dports="protocol",
                jump="ACCEPT",
            )
            == "Error: protocol must be specified"
        )

        assert (
            iptables.build_rule(
                "filter",
                "INPUT",
                command="I",
                position="3",
                full=True,
                sports="protocol",
                jump="ACCEPT",
            )
            == "Error: protocol must be specified"
        )

        assert (
            iptables.build_rule(
                "",
                "INPUT",
                command="I",
                position="3",
                full="True",
                match="state",
                jump="ACCEPT",
            )
            == "Error: Table needs to be specified"
        )

        assert (
            iptables.build_rule(
                "filter",
                "",
                command="I",
                position="3",
                full="True",
                match="state",
                jump="ACCEPT",
            )
            == "Error: Chain needs to be specified"
        )

        assert (
            iptables.build_rule(
                "filter",
                "INPUT",
                command="",
                position="3",
                full="True",
                match="state",
                jump="ACCEPT",
            )
            == "Error: Command needs to be specified"
        )

        # Test arguments that should appear after the --jump
        assert (
            iptables.build_rule(jump="REDIRECT", **{"to-port": 8080})
            == "--jump REDIRECT --to-port 8080"
        )

        # Should quote arguments with spaces, like log-prefix often has
        assert (
            iptables.build_rule(jump="LOG", **{"log-prefix": "long prefix"})
            == '--jump LOG --log-prefix "long prefix"'
        )

        # Should quote arguments with leading or trailing spaces
        assert (
            iptables.build_rule(jump="LOG", **{"log-prefix": "spam: "})
            == '--jump LOG --log-prefix "spam: "'
        )

        # Should allow no-arg jump options
        assert (
            iptables.build_rule(jump="CLUSTERIP", **{"new": ""})
            == "--jump CLUSTERIP --new"
        )

        # Should allow no-arg jump options as None
        assert (
            iptables.build_rule(jump="CT", **{"notrack": None}) == "--jump CT --notrack"
        )

        # should build match-sets with single string
        assert (
            iptables.build_rule(**{"match-set": "src flag1,flag2"})
            == "-m set --match-set src flag1,flag2"
        )

        # should build match-sets as list
        match_sets = [
            "src1 flag1",
            "src2 flag2,flag3",
        ]
        assert (
            iptables.build_rule(**{"match-set": match_sets})
            == "-m set --match-set src1 flag1 -m set --match-set src2 flag2,flag3"
        )

        # should handle negations for string match-sets
        assert (
            iptables.build_rule(**{"match-set": "!src flag"})
            == "-m set ! --match-set src flag"
        )

        # should handle negations for list match-sets
        match_sets = ["src1 flag", "not src2 flag2"]
        assert (
            iptables.build_rule(**{"match-set": match_sets})
            == "-m set --match-set src1 flag -m set ! --match-set src2 flag2"
        )

        # should allow escaped name
        assert (
            iptables.build_rule(**{"match": "recent", "name_": "SSH"})
            == "-m recent --name SSH"
        )

        # should allow empty arguments
        assert (
            iptables.build_rule(**{"match": "recent", "update": None})
            == "-m recent --update"
        )

        # Should allow the --save jump option to CONNSECMARK
        # assert (
        #    iptables.build_rule(jump='CONNSECMARK', **{'save': ''})
        #    == '--jump CONNSECMARK --save '
        # )

        ret = "/sbin/iptables --wait -t salt -I INPUT 3 -m state --jump ACCEPT"
        with patch.object(
            iptables, "_iptables_cmd", MagicMock(return_value="/sbin/iptables")
        ):
            assert (
                iptables.build_rule(
                    "salt",
                    "INPUT",
                    command="I",
                    position="3",
                    full="True",
                    match="state",
                    jump="ACCEPT",
                )
                == ret
            )


# 'get_saved_rules' function tests: 2


def test_get_saved_rules():
    """
    Test if it return a data structure of the rules in the conf file
    """
    mock = MagicMock(return_value=False)
    with patch.object(iptables, "_parse_conf", mock):
        assert not iptables.get_saved_rules()
        mock.assert_called_with(conf_file=None, family="ipv4")


def test_get_saved_rules_nilinuxrt():
    """
    Test get rules on NILinuxRT system
    """
    data = {
        "/etc/natinst/share/iptables.conf": textwrap.dedent(
            """\
            # Generated by iptables-save v1.6.2 on Thu Oct  8 12:13:44 2020
            *filter
            :INPUT ACCEPT [2958:584773]
            :FORWARD ACCEPT [0:0]
            :OUTPUT ACCEPT [92:23648]
            -A INPUT -p tcp -m tcp --dport 80 -j ACCEPT
            COMMIT
            # Completed on Thu Oct  8 12:13:44 2020
            """
        )
    }
    expected_input_rules = [
        {
            "protocol": ["tcp"],
            "jump": ["ACCEPT"],
            "match": ["tcp"],
            "destination_port": ["80"],
        }
    ]
    file_mock = mock_open(read_data=data)
    with patch.dict(iptables.__grains__, {"os_family": "NILinuxRT", "os": "NILinuxRT"}):
        with patch.object(iptables.salt.utils.files, "fopen", file_mock):
            rules = iptables.get_saved_rules()
            assert expected_input_rules == rules["filter"]["INPUT"]["rules"]


# 'get_rules' function tests: 1


def test_get_rules():
    """
    Test if it return a data structure of the current, in-memory rules
    """
    mock = MagicMock(return_value=False)
    with patch.object(iptables, "_parse_conf", mock):
        assert not iptables.get_rules()
        mock.assert_called_with(in_mem=True, family="ipv4")


# 'get_saved_policy' function tests: 1


def test_get_saved_policy():
    """
    Test if it return the current policy for the specified table/chain
    """
    assert (
        iptables.get_saved_policy(
            table="filter", chain=None, conf_file=None, family="ipv4"
        )
        == "Error: Chain needs to be specified"
    )

    with patch.object(
        iptables,
        "_parse_conf",
        MagicMock(return_value={"filter": {"INPUT": {"policy": True}}}),
    ):
        assert iptables.get_saved_policy(
            table="filter", chain="INPUT", conf_file=None, family="ipv4"
        )

    with patch.object(
        iptables,
        "_parse_conf",
        MagicMock(return_value={"filter": {"INPUT": {"policy1": True}}}),
    ):
        assert (
            iptables.get_saved_policy(
                table="filter", chain="INPUT", conf_file=None, family="ipv4"
            )
            is None
        )


# 'get_policy' function tests: 1


def test_get_policy():
    """
    Test if it return the current policy for the specified table/chain
    """
    assert (
        iptables.get_policy(table="filter", chain=None, family="ipv4")
        == "Error: Chain needs to be specified"
    )

    with patch.object(
        iptables,
        "_parse_conf",
        MagicMock(return_value={"filter": {"INPUT": {"policy": True}}}),
    ):
        assert iptables.get_policy(table="filter", chain="INPUT", family="ipv4")

    with patch.object(
        iptables,
        "_parse_conf",
        MagicMock(return_value={"filter": {"INPUT": {"policy1": True}}}),
    ):
        assert iptables.get_policy(table="filter", chain="INPUT", family="ipv4") is None


# 'set_policy' function tests: 1


def test_set_policy():
    """
    Test if it set the current policy for the specified table/chain
    """
    with patch.object(iptables, "_has_option", MagicMock(return_value=True)):
        assert (
            iptables.set_policy(table="filter", chain=None, policy=None, family="ipv4")
            == "Error: Chain needs to be specified"
        )

        assert (
            iptables.set_policy(
                table="filter", chain="INPUT", policy=None, family="ipv4"
            )
            == "Error: Policy needs to be specified"
        )

        mock = MagicMock(return_value=True)
        with patch.dict(iptables.__salt__, {"cmd.run_stderr": mock}):
            assert iptables.set_policy(
                table="filter", chain="INPUT", policy="ACCEPT", family="ipv4"
            )


# 'save' function tests: 1


def test_save():
    """
    Test if it save the current in-memory rules to disk
    """
    with patch("salt.modules.iptables._conf", MagicMock(return_value=False)), patch(
        "os.path.isdir", MagicMock(return_value=True)
    ):
        mock = MagicMock(return_value=True)
        with patch.dict(
            iptables.__salt__,
            {
                "cmd.run_stdout": mock,
                "file.write": mock,
                "config.option": MagicMock(return_value=[]),
            },
        ):
            assert iptables.save(filename="/xyz", family="ipv4")


# 'check' function tests: 1


def test_check():
    """
    Test if it check for the existence of a rule in the table and chain
    """
    assert (
        iptables.check(table="filter", chain=None, rule=None, family="ipv4")
        == "Error: Chain needs to be specified"
    )

    assert (
        iptables.check(table="filter", chain="INPUT", rule=None, family="ipv4")
        == "Error: Rule needs to be specified"
    )

    mock_rule = "m state --state RELATED,ESTABLISHED -j ACCEPT"
    mock_chain = "INPUT"
    mock_uuid = 31337
    mock_cmd_rule = MagicMock(return_value=f"-A {mock_chain}\n-A {hex(mock_uuid)}")
    mock_cmd_nooutput = MagicMock(return_value="")
    mock_has = MagicMock(return_value=True)
    mock_not = MagicMock(return_value=False)

    with patch.object(iptables, "_has_option", mock_not):
        with patch.object(uuid, "getnode", MagicMock(return_value=mock_uuid)):
            with patch.dict(
                iptables.__salt__,
                {"cmd.run_stdout": mock_cmd_rule, "cmd.run": mock_cmd_nooutput},
            ):
                assert iptables.check(
                    table="filter",
                    chain=mock_chain,
                    rule=mock_rule,
                    family="ipv4",
                )

    mock_cmd = MagicMock(return_value="")

    with patch.object(iptables, "_has_option", mock_not):
        with patch.object(uuid, "getnode", MagicMock(return_value=mock_uuid)):
            with patch.dict(
                iptables.__salt__,
                {
                    "cmd.run_stdout": mock_cmd_nooutput,
                    "cmd.run": mock_cmd_nooutput,
                },
            ):
                assert not iptables.check(
                    table="filter",
                    chain=mock_chain,
                    rule=mock_rule,
                    family="ipv4",
                )

    with patch.object(iptables, "_has_option", mock_has):
        with patch.dict(iptables.__salt__, {"cmd.run_stderr": mock_cmd}):
            with patch.dict(iptables.__context__, {"retcode": 1}):
                assert not iptables.check(
                    table="filter", chain="INPUT", rule=mock_rule, family="ipv4"
                )

    mock_cmd = MagicMock(return_value="-A 0x4d2")
    mock_uuid = MagicMock(return_value=1234)

    with patch.object(iptables, "_has_option", mock_has):
        with patch.object(uuid, "getnode", mock_uuid):
            with patch.dict(iptables.__salt__, {"cmd.run_stderr": mock_cmd}):
                with patch.dict(iptables.__context__, {"retcode": 0}):
                    assert iptables.check(
                        table="filter", chain="0x4d2", rule=mock_rule, family="ipv4"
                    )


# 'check_chain' function tests: 1


def test_check_chain():
    """
    Test if it check for the existence of a chain in the table
    """
    assert (
        iptables.check_chain(table="filter", chain=None, family="ipv4")
        == "Error: Chain needs to be specified"
    )

    mock_cmd = MagicMock(return_value="")
    with patch.dict(iptables.__salt__, {"cmd.run_stdout": mock_cmd}):
        assert not iptables.check_chain(table="filter", chain="INPUT", family="ipv4")


# 'new_chain' function tests: 1


def test_new_chain():
    """
    Test if it create new custom chain to the specified table.
    """
    assert (
        iptables.new_chain(table="filter", chain=None, family="ipv4")
        == "Error: Chain needs to be specified"
    )

    mock_cmd = MagicMock(return_value="")
    with patch.dict(
        iptables.__salt__,
        {
            "cmd.run_stdout": mock_cmd,  # called by iptables._has_option
            "cmd.run_stderr": mock_cmd,
        },
    ):
        assert iptables.new_chain(table="filter", chain="INPUT", family="ipv4")


# 'delete_chain' function tests: 1


def test_delete_chain():
    """
    Test if it delete custom chain to the specified table.
    """
    assert (
        iptables.delete_chain(table="filter", chain=None, family="ipv4")
        == "Error: Chain needs to be specified"
    )

    mock_cmd = MagicMock(return_value="")
    with patch.dict(
        iptables.__salt__,
        {
            "cmd.run_stdout": mock_cmd,  # called by iptables._has_option
            "cmd.run_stderr": mock_cmd,
        },
    ):
        assert iptables.delete_chain(table="filter", chain="INPUT", family="ipv4")


# 'append' function tests: 1


def test_append():
    """
    Test if it append a rule to the specified table/chain.
    """
    with patch.object(
        iptables, "_has_option", MagicMock(return_value=True)
    ), patch.object(iptables, "check", MagicMock(return_value=False)):
        assert (
            iptables.append(table="filter", chain=None, rule=None, family="ipv4")
            == "Error: Chain needs to be specified"
        )

        assert (
            iptables.append(table="filter", chain="INPUT", rule=None, family="ipv4")
            == "Error: Rule needs to be specified"
        )

        _rule = "m state --state RELATED,ESTABLISHED -j ACCEPT"
        mock = MagicMock(side_effect=["", "SALT"])
        with patch.dict(iptables.__salt__, {"cmd.run_stderr": mock}):
            assert iptables.append(
                table="filter", chain="INPUT", rule=_rule, family="ipv4"
            )

            assert not iptables.append(
                table="filter", chain="INPUT", rule=_rule, family="ipv4"
            )


# 'insert' function tests: 1


def test_insert():
    """
    Test if it insert a rule into the specified table/chain,
    at the specified position.
    """
    with patch.object(
        iptables, "_has_option", MagicMock(return_value=True)
    ), patch.object(iptables, "check", MagicMock(return_value=False)):
        assert (
            iptables.insert(
                table="filter", chain=None, position=None, rule=None, family="ipv4"
            )
            == "Error: Chain needs to be specified"
        )

        pos_err = "Error: Position needs to be specified or use append (-A)"
        assert (
            iptables.insert(
                table="filter",
                chain="INPUT",
                position=None,
                rule=None,
                family="ipv4",
            )
            == pos_err
        )

        assert (
            iptables.insert(
                table="filter", chain="INPUT", position=3, rule=None, family="ipv4"
            )
            == "Error: Rule needs to be specified"
        )

        _rule = "m state --state RELATED,ESTABLISHED -j ACCEPT"
        mock = MagicMock(return_value=True)
        with patch.dict(iptables.__salt__, {"cmd.run_stderr": mock}):
            assert iptables.insert(
                table="filter",
                chain="INPUT",
                position=3,
                rule=_rule,
                family="ipv4",
            )


# 'delete' function tests: 1


def test_delete():
    """
    Test if it delete a rule from the specified table/chain
    """
    with patch.object(iptables, "_has_option", MagicMock(return_value=True)):
        _rule = "m state --state RELATED,ESTABLISHED -j ACCEPT"
        assert (
            iptables.delete(
                table="filter", chain=None, position=3, rule=_rule, family="ipv4"
            )
            == "Error: Only specify a position or a rule, not both"
        )

        mock = MagicMock(return_value=True)
        with patch.dict(iptables.__salt__, {"cmd.run_stderr": mock}):
            assert iptables.delete(
                table="filter",
                chain="INPUT",
                position=3,
                rule="",
                family="ipv4",
            )


# 'flush' function tests: 1


def test_flush():
    """
    Test if it flush the chain in the specified table,
    flush all chains in the specified table if not specified chain.
    """
    with patch.object(iptables, "_has_option", MagicMock(return_value=True)):
        mock_cmd = MagicMock(return_value=True)
        with patch.dict(iptables.__salt__, {"cmd.run_stderr": mock_cmd}):
            assert iptables.flush(table="filter", chain="INPUT", family="ipv4")



import sys

import pytest

import salt.utils.event
import salt.utils.platform
import tests.support.helpers
from tests.conftest import FIPS_TESTRUN


@pytest.fixture
def salt_master_1(request, salt_factories):
    config_defaults = {
        "open_mode": True,
        "transport": request.config.getoption("--transport"),
    }
    config_overrides = {
        "interface": "127.0.0.1",
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }

    factory = salt_factories.salt_master_daemon(
        "master-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture
def salt_minion_1(salt_master_1):
    config_defaults = {
        "transport": salt_master_1.config["transport"],
    }
    master_1_port = salt_master_1.config["ret_port"]
    master_1_addr = salt_master_1.config["interface"]
    config_overrides = {
        "master": [
            f"{master_1_addr}:{master_1_port}",
        ],
        "test.foo": "baz",
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = salt_master_1.salt_minion_daemon(
        "minion-1",
        defaults=config_defaults,
        overrides=config_overrides,
        extra_cli_arguments_after_first_start_failure=["--log-level=info"],
    )
    with factory.started(start_timeout=120):
        yield factory


@pytest.fixture
def script(salt_minion_1, tmp_path):
    path = tmp_path / "script.py"
    content = f"""
    import salt.config
    import salt.utils.event

    opts = salt.config.minion_config('{salt_minion_1.config_file}')

    payload =  b'0' * 1048576000

    big_event = dict()
    for i in range(10000):
        big_event[i] = payload =  b'0' * 100

    with salt.utils.event.get_event("minion", opts=opts) as event:
        event.fire_master(big_event, 'bigevent')

    """
    path.write_text(tests.support.helpers.dedent(content))
    return path


# @pytest.mark.timeout_unless_on_windows(360)
def test_schedule_large_event(salt_master_1, salt_minion_1, script):
    cli = salt_master_1.salt_cli(timeout=120)
    ret = cli.run(
        "schedule.add",
        name="myjob",
        function="cmd.run",
        seconds=5,
        job_args=f'["{sys.executable} {script}"]',
        minion_tgt=salt_minion_1.id,
    )
    assert "result" in ret.data
    assert ret.data["result"]
    with salt.utils.event.get_event(
        "master",
        salt_master_1.config["sock_dir"],
        salt_master_1.config,
    ) as event:
        event = event.get_event(tag="bigevent", wait=15)
        assert event
        assert "data" in event
        assert len(event["data"]) == 10000



"""
Interface to Red Hat tuned-adm module

:maintainer:    Syed Ali <alicsyed@gmail.com>
:maturity:      new
:depends:       cmd.run
:platform:      Linux
"""

import salt.exceptions
from salt.modules.tuned import TUNED_OFF_RETURN_NAME


def profile(name):
    """
    This state module allows you to modify system tuned parameters

    Example tuned.sls file to set profile to virtual-guest

    tuned:
      tuned.profile
        - name: virtual-guest

    name
        tuned profile name to set the system to

    To see a valid list of states call execution module:
        :py:func:`tuned.list <salt.modules.tuned.list_>`
    """

    # create data-structure to return with default value
    ret = {"name": "", "changes": {}, "result": False, "comment": ""}

    ret[name] = name
    profile = name

    # get the current state of tuned-adm
    current_state_dict = __salt__["tuned.active"]()

    # Off is returned as retcode 1, stdout == TUNED_OFF_RETURN_NAME
    # any other type of error will be returned here
    if (
        current_state_dict["retcode"] != 0
        and current_state_dict["stdout"] != TUNED_OFF_RETURN_NAME
    ):
        ret["comment"] = current_state_dict["stderr"]
        return ret

    # if current state is same as requested state, return without doing much
    if current_state_dict["retcode"] == 0 and profile == current_state_dict["stdout"]:
        ret["result"] = True
        ret["comment"] = "System already in the correct state"
        return ret

    # test mode
    if __opts__["test"] is True:
        # Only perform the valid profile test if it is a test,
        #   tuned-adm command will fail and return message
        valid_profiles = __salt__["tuned.list"]()
        if profile not in valid_profiles:
            raise salt.exceptions.SaltInvocationError("Invalid Profile Name")
        ret["comment"] = 'The state of "{}" will be changed.'.format(
            current_state_dict["stdout"]
        )
        ret["changes"] = {
            "old": current_state_dict["stdout"],
            "new": f"Profile will be set to {profile}",
        }
        # return None when testing
        ret["result"] = None
        return ret

    # we come to this stage if current state was determined and is different that requested state
    # we there have to set the new state request
    new_state_dict = __salt__["tuned.profile"](profile)

    if new_state_dict["retcode"] != 0:
        ret["comment"] = new_state_dict["stderr"]
    else:
        # create the comment data structure
        ret["comment"] = f'Tunings changed to "{profile}"'
        # add the changes specifics
        ret["changes"] = {
            "old": current_state_dict["stdout"],
            "new": new_state_dict["stdout"],
        }
        ret["result"] = True

    # return with the dictionary data structure
    return ret


def off(name=None):
    """

    Turns 'tuned' off.
    Example tuned.sls file for turning tuned off:

    tuned:
      tuned.off: []


    To see a valid list of states call execution module:
        :py:func:`tuned.list <salt.modules.tuned.list_>`
    """

    # create data-structure to return with default value
    ret = {"name": "off", "changes": {}, "result": False, "comment": "off"}

    # check the current state of tuned
    current_state_dict = __salt__["tuned.active"]()

    # Off is returned as retcode 1, stdout == TUNED_OFF_RETURN_NAME
    if current_state_dict["retcode"] != 0:
        if current_state_dict["stdout"] == TUNED_OFF_RETURN_NAME:
            ret["result"] = True
            ret["comment"] = "System already in the correct state"
            return ret
        ret["comment"] = current_state_dict["stderr"]
        return ret

    # test mode
    if __opts__["test"] is True:
        ret["comment"] = 'The state of "{}" will be turned off.'.format(
            current_state_dict["stdout"]
        )
        ret["changes"] = {
            "old": current_state_dict["stdout"],
            "new": "Profile will be set to off",
        }
        # return None when testing
        ret["result"] = None
        return ret

    # execute the tuned.off module
    off_result_dict = __salt__["tuned.off"]()

    if off_result_dict["retcode"] != 0:
        ret["comment"] = off_result_dict["stderr"]
    else:
        ret["comment"] = "Tunings have been turned off"
        ret["changes"] = {
            "old": current_state_dict["stdout"],
            "new": "off",
        }
        ret["result"] = True

    # return with the dictionary data structure
    return ret



import pytest

import salt.modules.debian_service as debian_service
from tests.support.mock import DEFAULT, MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {debian_service: {}}


def test_get_enabled():
    init_d_globs = ["/etc/init.d/S50foo", "/etc/init.d/S90bar"]
    glob_mock = MagicMock(
        side_effect=lambda x: init_d_globs if x == "/etc/rc[S3].d/S*" else DEFAULT
    )
    with patch("glob.glob", glob_mock), patch.object(
        debian_service, "_get_runlevel", MagicMock(return_value="3")
    ):
        ret = debian_service.get_enabled()
        expected = ["bar", "foo"]
        assert ret == expected, ret


def test_get_disabled():
    get_all = MagicMock(return_value=["foo", "bar", "baz"])
    get_enabled = MagicMock(return_value=["bar", "baz"])
    with patch.object(debian_service, "get_all", get_all), patch.object(
        debian_service, "get_enabled", get_enabled
    ):
        ret = debian_service.get_disabled()
        expected = ["foo"]
        assert ret == expected, ret


def test_available():
    get_all = MagicMock(return_value=["foo", "bar", "baz"])
    with patch.object(debian_service, "get_all", get_all):
        assert not debian_service.available("qux")
        assert debian_service.available("foo")


def test_missing():
    get_all = MagicMock(return_value=["foo", "bar", "baz"])
    with patch.object(debian_service, "get_all", get_all):
        assert debian_service.missing("qux")
        assert not debian_service.missing("foo")


def test_get_all():
    get_enabled = MagicMock(return_value=["baz", "hello", "world"])
    init_d_globs = ["/etc/init.d/foo", "/etc/init.d/bar"]
    glob_mock = MagicMock(
        side_effect=lambda x: init_d_globs if x == "/etc/init.d/*" else DEFAULT
    )
    with patch("glob.glob", glob_mock), patch.object(
        debian_service, "get_enabled", get_enabled
    ):
        ret = debian_service.get_all()
        expected = ["bar", "baz", "foo", "hello", "world"]
        assert ret == expected, ret


def test_start():
    mock = MagicMock(return_value=0)
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
        # Test successful command (0 retcode)
        assert debian_service.start("foo")
        # Confirm expected command was run
        mock.assert_called_once_with("service foo start")
        # Test unsuccessful command (nonzero retcode)
        mock.return_value = 1
        assert not debian_service.start("foo")


def test_stop():
    mock = MagicMock(return_value=0)
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
        # Test successful command (0 retcode)
        assert debian_service.stop("foo")
        # Confirm expected command was run
        mock.assert_called_once_with("service foo stop")
        # Test unsuccessful command (nonzero retcode)
        mock.return_value = 1
        assert not debian_service.stop("foo")


def test_restart():
    mock = MagicMock(return_value=0)
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
        # Test successful command (0 retcode)
        assert debian_service.restart("foo")
        # Confirm expected command was run
        mock.assert_called_once_with("service foo restart")
        # Test unsuccessful command (nonzero retcode)
        mock.return_value = 1
        assert not debian_service.restart("foo")


def test_reload_():
    mock = MagicMock(return_value=0)
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
        # Test successful command (0 retcode)
        assert debian_service.reload_("foo")
        # Confirm expected command was run
        mock.assert_called_once_with("service foo reload")
        # Test unsuccessful command (nonzero retcode)
        mock.return_value = 1
        assert not debian_service.reload_("foo")


def test_force_reload():
    mock = MagicMock(return_value=0)
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
        # Test successful command (0 retcode)
        assert debian_service.force_reload("foo")
        # Confirm expected command was run
        mock.assert_called_once_with("service foo force-reload")
        # Test unsuccessful command (nonzero retcode)
        mock.return_value = 1
        assert not debian_service.force_reload("foo")


def test_status():
    mock = MagicMock(return_value="123")
    with patch.dict(debian_service.__salt__, {"status.pid": mock}):
        assert debian_service.status("foo", "foobar")

    mock = MagicMock(return_value=0)
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
        # Test successful command (0 retcode)
        assert debian_service.status("foo")
        # Confirm expected command was run
        mock.assert_called_once_with("service foo status", ignore_retcode=True)
        # Test unsuccessful command (nonzero retcode)
        mock.return_value = 1
        assert not debian_service.enable("foo")

    mock = MagicMock(side_effect=lambda x, **y: 0 if x == "service bar status" else 1)
    get_all = MagicMock(return_value=["foo", "bar", "baz"])
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}), patch.object(
        debian_service, "get_all", get_all
    ):
        ret = debian_service.status("b*")
        expected = {"bar": True, "baz": False}
        assert ret == expected, ret


def test_enable():
    mock = MagicMock(return_value=0)
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
        # Test successful command (0 retcode)
        assert debian_service.enable("foo")
        # Confirm expected command was run
        mock.assert_called_once_with(
            "insserv foo && update-rc.d foo enable", python_shell=True
        )
        # Test unsuccessful command (nonzero retcode)
        mock.return_value = 1
        assert not debian_service.enable("foo")


def test_disable():
    mock = MagicMock(return_value=0)
    with patch.dict(debian_service.__salt__, {"cmd.retcode": mock}):
        # Test successful command (0 retcode)
        assert debian_service.disable("foo")
        # Confirm expected command was run
        mock.assert_called_once_with("update-rc.d foo disable")
        # Test unsuccessful command (nonzero retcode)
        mock.return_value = 1
        assert not debian_service.disable("foo")


def test_enabled():
    mock = MagicMock(return_value=["foo"])
    with patch.object(debian_service, "get_enabled", mock):
        assert debian_service.enabled("foo")
        assert not debian_service.enabled("bar")


def test_disabled():
    mock = MagicMock(return_value=["foo"])
    with patch.object(debian_service, "get_disabled", mock):
        assert debian_service.disabled("foo")
        assert not debian_service.disabled("bar")



"""
    :codeauthor: :email:`Alexandru Bleotu <alexandru.bleotu@morganstanley.com>`

    Tests for functions in salt.states.esxdatacenter
"""

import salt.states.esxdatacenter as esxdatacenter
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class DatacenterConfiguredTestCase(TestCase, LoaderModuleMockMixin):
    """Tests for salt.modules.esxdatacenter.datacenter_configured"""

    def setup_loader_modules(self):
        return {esxdatacenter: {}}

    def setUp(self):
        self.mock_si = MagicMock()
        self.mock_dc = MagicMock()

        patcher = patch.dict(
            esxdatacenter.__salt__,
            {
                "vsphere.get_proxy_type": MagicMock(),
                "vsphere.get_service_instance_via_proxy": MagicMock(
                    return_value=self.mock_si
                ),
                "vsphere.list_datacenters_via_proxy": MagicMock(
                    return_value=[self.mock_dc]
                ),
                "vsphere.disconnect": MagicMock(),
            },
        )
        patcher.start()
        self.addCleanup(patcher.stop)
        patcher = patch.dict(esxdatacenter.__opts__, {"test": False})
        patcher.start()
        self.addCleanup(patcher.stop)

    def tearDown(self):
        for attrname in ("mock_si",):
            try:
                delattr(self, attrname)
            except AttributeError:
                continue

    def test_dc_name_different_proxy(self):
        with patch.dict(
            esxdatacenter.__salt__,
            {"vsphere.get_proxy_type": MagicMock(return_value="different_proxy")},
        ):
            res = esxdatacenter.datacenter_configured("fake_dc")
        self.assertDictEqual(
            res,
            {
                "name": "fake_dc",
                "changes": {},
                "result": True,
                "comment": "Datacenter 'fake_dc' already exists. Nothing to be done.",
            },
        )

    def test_dc_name_esxdatacenter_proxy(self):
        with patch.dict(
            esxdatacenter.__salt__,
            {
                "vsphere.get_proxy_type": MagicMock(return_value="esxdatacenter"),
                "esxdatacenter.get_details": MagicMock(
                    return_value={"datacenter": "proxy_dc"}
                ),
            },
        ):
            res = esxdatacenter.datacenter_configured("fake_dc")
        self.assertDictEqual(
            res,
            {
                "name": "fake_dc",
                "changes": {},
                "result": True,
                "comment": "Datacenter 'proxy_dc' already exists. Nothing to be done.",
            },
        )

    def test_get_service_instance(self):
        mock_get_service_instance = MagicMock()
        with patch.dict(
            esxdatacenter.__salt__,
            {"vsphere.get_service_instance_via_proxy": mock_get_service_instance},
        ):
            esxdatacenter.datacenter_configured("fake_dc")
        mock_get_service_instance.assert_called_once_with()

    def test_list_datacenters(self):
        mock_list_datacenters = MagicMock()
        with patch.dict(
            esxdatacenter.__salt__,
            {"vsphere.list_datacenters_via_proxy": mock_list_datacenters},
        ):
            esxdatacenter.datacenter_configured("fake_dc")
        mock_list_datacenters.assert_called_once_with(
            datacenter_names=["fake_dc"], service_instance=self.mock_si
        )

    def test_create_datacenter(self):
        mock_create_datacenter = MagicMock()
        with patch.dict(
            esxdatacenter.__salt__,
            {
                "vsphere.list_datacenters_via_proxy": MagicMock(return_value=[]),
                "vsphere.create_datacenter": mock_create_datacenter,
            },
        ):
            res = esxdatacenter.datacenter_configured("fake_dc")
        mock_create_datacenter.assert_called_once_with("fake_dc", self.mock_si)
        self.assertDictEqual(
            res,
            {
                "name": "fake_dc",
                "changes": {"new": {"name": "fake_dc"}},
                "result": True,
                "comment": "Created datacenter 'fake_dc'.",
            },
        )

    def test_create_datacenter_test_mode(self):
        with patch.dict(esxdatacenter.__opts__, {"test": True}):
            with patch.dict(
                esxdatacenter.__salt__,
                {"vsphere.list_datacenters_via_proxy": MagicMock(return_value=[])},
            ):
                res = esxdatacenter.datacenter_configured("fake_dc")
        self.assertDictEqual(
            res,
            {
                "name": "fake_dc",
                "changes": {"new": {"name": "fake_dc"}},
                "result": None,
                "comment": "State will create datacenter 'fake_dc'.",
            },
        )

    def test_nothing_to_be_done_test_mode(self):
        with patch.dict(esxdatacenter.__opts__, {"test": True}):
            with patch.dict(
                esxdatacenter.__salt__,
                {"vsphere.get_proxy_type": MagicMock(return_value="different_proxy")},
            ):
                res = esxdatacenter.datacenter_configured("fake_dc")
        self.assertDictEqual(
            res,
            {
                "name": "fake_dc",
                "changes": {},
                "result": True,
                "comment": "Datacenter 'fake_dc' already exists. Nothing to be done.",
            },
        )

    def test_state_get_service_instance_raise_command_execution_error(self):
        mock_disconnect = MagicMock()
        with patch.dict(
            esxdatacenter.__salt__,
            {
                "vsphere.disconnect": mock_disconnect,
                "vsphere.get_service_instance_via_proxy": MagicMock(
                    side_effect=CommandExecutionError("Error")
                ),
            },
        ):
            res = esxdatacenter.datacenter_configured("fake_dc")
        self.assertEqual(mock_disconnect.call_count, 0)
        self.assertDictEqual(
            res, {"name": "fake_dc", "changes": {}, "result": False, "comment": "Error"}
        )

    def test_state_raise_command_execution_error_after_si(self):
        mock_disconnect = MagicMock()
        with patch.dict(
            esxdatacenter.__salt__,
            {
                "vsphere.disconnect": mock_disconnect,
                "vsphere.list_datacenters_via_proxy": MagicMock(
                    side_effect=CommandExecutionError("Error")
                ),
            },
        ):
            res = esxdatacenter.datacenter_configured("fake_dc")
        mock_disconnect.assert_called_once_with(self.mock_si)
        self.assertDictEqual(
            res, {"name": "fake_dc", "changes": {}, "result": False, "comment": "Error"}
        )

    def test_state_raise_command_execution_error_test_mode(self):
        with patch.dict(esxdatacenter.__opts__, {"test": True}):
            with patch.dict(
                esxdatacenter.__salt__,
                {
                    "vsphere.list_datacenters_via_proxy": MagicMock(
                        side_effect=CommandExecutionError("Error")
                    )
                },
            ):
                res = esxdatacenter.datacenter_configured("fake_dc")
        self.assertDictEqual(
            res, {"name": "fake_dc", "changes": {}, "result": None, "comment": "Error"}
        )



import pytest

import salt.modules.macpackage as macpackage
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {macpackage: {}}


def test_install():
    """
    Test installing a PKG file
    """
    mock = MagicMock()
    with patch.dict(macpackage.__salt__, {"cmd.run_all": mock}):
        macpackage.install("/path/to/file.pkg")
        mock.assert_called_once_with(
            "installer -pkg /path/to/file.pkg -target LocalSystem",
            python_shell=False,
        )


def test_install_wildcard():
    """
    Test installing a PKG file with a wildcard
    """
    mock = MagicMock()
    with patch.dict(macpackage.__salt__, {"cmd.run_all": mock}):
        macpackage.install("/path/to/*.pkg")
        mock.assert_called_once_with(
            "installer -pkg /path/to/*.pkg -target LocalSystem", python_shell=True
        )


def test_install_with_extras():
    """
    Test installing a PKG file with extra options
    """
    mock = MagicMock()
    with patch.dict(macpackage.__salt__, {"cmd.run_all": mock}):
        macpackage.install("/path/to/file.pkg", store=True, allow_untrusted=True)
        mock.assert_called_once_with(
            "installer -pkg /path/to/file.pkg -target LocalSystem -store"
            " -allowUntrusted",
            python_shell=False,
        )


def test_install_app():
    """
    Test installing an APP package
    """
    mock = MagicMock()
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        macpackage.install_app("/path/to/file.app")
        mock.assert_called_once_with(
            'rsync -a --delete "/path/to/file.app/" "/Applications/file.app"'
        )


def test_install_app_specify_target():
    """
    Test installing an APP package with a specific target
    """
    mock = MagicMock()
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        macpackage.install_app("/path/to/file.app", "/Applications/new.app")
        mock.assert_called_once_with(
            'rsync -a --delete "/path/to/file.app/" "/Applications/new.app"'
        )


def test_install_app_with_slash():
    """
    Test installing an APP package with a specific target
    """
    mock = MagicMock()
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        macpackage.install_app("/path/to/file.app/")
        mock.assert_called_once_with(
            'rsync -a --delete "/path/to/file.app/" "/Applications/file.app"'
        )


def test_uninstall():
    """
    Test Uninstalling an APP package with a specific target
    """
    mock = MagicMock()
    with patch.dict(macpackage.__salt__, {"file.remove": mock}):
        macpackage.uninstall_app("/path/to/file.app")
        mock.assert_called_once_with("/path/to/file.app")


def test_mount():
    """
    Test mounting an dmg file to a temporary location
    """
    cmd_mock = MagicMock()
    temp_mock = MagicMock(return_value="dmg-ABCDEF")
    with patch.dict(macpackage.__salt__, {"cmd.run": cmd_mock, "temp.dir": temp_mock}):
        macpackage.mount("/path/to/file.dmg")
        temp_mock.assert_called_once_with(prefix="dmg-")
        cmd_mock.assert_called_once_with(
            "hdiutil attach -readonly -nobrowse -mountpoint "
            'dmg-ABCDEF "/path/to/file.dmg"'
        )


def test_unmount():
    """
    Test Unmounting an dmg file to a temporary location
    """
    mock = MagicMock()
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        macpackage.unmount("/path/to/file.dmg")
        mock.assert_called_once_with('hdiutil detach "/path/to/file.dmg"')


def test_installed_pkgs():
    """
    Test getting a list of the installed packages
    """
    expected = ["com.apple.this", "com.salt.that"]
    mock = MagicMock(return_value="com.apple.this\ncom.salt.that")
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        out = macpackage.installed_pkgs()
        mock.assert_called_once_with("pkgutil --pkgs")
        assert out == expected


def test_get_pkg_id_with_files():
    """
    Test getting a the id for a package
    """
    with patch(
        "salt.modules.macpackage._get_pkg_id_from_pkginfo"
    ) as pkg_id_pkginfo_mock:
        expected = ["com.apple.this"]
        cmd_mock = MagicMock(
            side_effect=[
                "/path/to/PackageInfo\n/path/to/some/other/fake/PackageInfo",
                "",
                "",
            ]
        )
        pkg_id_pkginfo_mock.side_effect = [["com.apple.this"], []]
        temp_mock = MagicMock(return_value="/tmp/dmg-ABCDEF")
        remove_mock = MagicMock()

        with patch.dict(
            macpackage.__salt__,
            {
                "cmd.run": cmd_mock,
                "temp.dir": temp_mock,
                "file.remove": remove_mock,
            },
        ):
            out = macpackage.get_pkg_id("/path/to/file.pkg")

            temp_mock.assert_called_once_with(prefix="pkg-")
            cmd_calls = [
                call(
                    "xar -t -f /path/to/file.pkg | grep PackageInfo",
                    python_shell=True,
                    output_loglevel="quiet",
                ),
                call(
                    "xar -x -f /path/to/file.pkg /path/to/PackageInfo"
                    " /path/to/some/other/fake/PackageInfo",
                    cwd="/tmp/dmg-ABCDEF",
                    output_loglevel="quiet",
                ),
            ]
            cmd_mock.assert_has_calls(cmd_calls)

            pkg_id_pkginfo_calls = [
                call("/path/to/PackageInfo"),
                call("/path/to/some/other/fake/PackageInfo"),
            ]
            pkg_id_pkginfo_mock.assert_has_calls(pkg_id_pkginfo_calls)
            remove_mock.assert_called_once_with("/tmp/dmg-ABCDEF")

            assert out == expected


def test_get_pkg_id_with_dir():
    """
    Test getting a the id for a package with a directory
    """
    with patch("salt.modules.macpackage._get_pkg_id_dir") as pkg_id_dir_mock:
        expected = ["com.apple.this"]
        pkg_id_dir_mock.return_value = ["com.apple.this"]
        cmd_mock = MagicMock(return_value="Error opening /path/to/file.pkg")
        temp_mock = MagicMock(return_value="/tmp/dmg-ABCDEF")
        remove_mock = MagicMock()

        with patch.dict(
            macpackage.__salt__,
            {
                "cmd.run": cmd_mock,
                "temp.dir": temp_mock,
                "file.remove": remove_mock,
            },
        ):
            out = macpackage.get_pkg_id("/path/to/file.pkg")

            temp_mock.assert_called_once_with(prefix="pkg-")
            cmd_mock.assert_called_once_with(
                "xar -t -f /path/to/file.pkg | grep PackageInfo",
                python_shell=True,
                output_loglevel="quiet",
            )
            pkg_id_dir_mock.assert_called_once_with("/path/to/file.pkg")
            remove_mock.assert_called_once_with("/tmp/dmg-ABCDEF")

            assert out == expected


def test_get_mpkg_ids():
    """
    Test getting the ids of a mpkg file
    """
    with patch("salt.modules.macpackage.get_pkg_id") as get_pkg_id_mock:
        expected = ["com.apple.this", "com.salt.other"]
        mock = MagicMock(return_value="/tmp/dmg-X/file.pkg\n/tmp/dmg-X/other.pkg")
        get_pkg_id_mock.side_effect = [["com.apple.this"], ["com.salt.other"]]

        with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
            out = macpackage.get_mpkg_ids("/path/to/file.mpkg")

            mock.assert_called_once_with("find /path/to -name *.pkg", python_shell=True)

            calls = [call("/tmp/dmg-X/file.pkg"), call("/tmp/dmg-X/other.pkg")]
            get_pkg_id_mock.assert_has_calls(calls)

            assert out == expected


def test_get_pkg_id_from_pkginfo():
    """
    Test getting a package id from pkginfo files
    """
    expected = ["com.apple.this", "com.apple.that"]
    mock = MagicMock(return_value="com.apple.this\ncom.apple.that")
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        out = macpackage._get_pkg_id_from_pkginfo("/tmp/dmg-X/PackageInfo")
        cmd = (
            "cat /tmp/dmg-X/PackageInfo | grep -Eo"
            " 'identifier=\"[a-zA-Z.0-9\\-]*\"' | cut -c 13- | tr -d '\"'"
        )
        mock.assert_called_once_with(cmd, python_shell=True)
        assert out == expected


def test_get_pkg_id_from_pkginfo_no_file():
    """
    Test getting a package id from pkginfo file when it doesn't exist
    """
    expected = []
    mock = MagicMock(return_value="No such file")
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        out = macpackage._get_pkg_id_from_pkginfo("/tmp/dmg-X/PackageInfo")
        cmd = (
            "cat /tmp/dmg-X/PackageInfo | grep -Eo"
            " 'identifier=\"[a-zA-Z.0-9\\-]*\"' | cut -c 13- | tr -d '\"'"
        )
        mock.assert_called_once_with(cmd, python_shell=True)
        assert out == expected


def test_get_pkg_id_dir():
    """
    Test getting a package id from a directory
    """
    expected = ["com.apple.this"]
    mock = MagicMock(return_value="com.apple.this")
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        out = macpackage._get_pkg_id_dir("/tmp/dmg-X/")
        cmd = (
            '/usr/libexec/PlistBuddy -c "print :CFBundleIdentifier"'
            " /tmp/dmg-X/Contents/Info.plist"
        )
        mock.assert_called_once_with(cmd, python_shell=False)
        assert out == expected


def test_get_pkg_id_dir_wildcard():
    """
    Test getting a package id from a directory with a wildcard
    """
    expected = ["com.apple.this"]
    mock = MagicMock(return_value="com.apple.this")
    with patch.dict(macpackage.__salt__, {"cmd.run": mock}):
        out = macpackage._get_pkg_id_dir("/tmp/dmg-X/*.pkg/")
        cmd = (
            '/usr/libexec/PlistBuddy -c "print :CFBundleIdentifier"'
            " '/tmp/dmg-X/*.pkg/Contents/Info.plist'"
        )
        mock.assert_called_once_with(cmd, python_shell=True)
        assert out == expected



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import salt.states.group as group
from salt.utils.odict import OrderedDict
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class GroupTestCase(TestCase, LoaderModuleMockMixin):
    """
    Validate the group state
    """

    def setup_loader_modules(self):
        return {group: {}}

    def test_present(self):
        """
        Test to ensure that a group is present
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": {}}

        ret.update(
            {
                "comment": (
                    'Error: Conflicting options "members" with'
                    ' "addusers" and/or "delusers" can not be used together. '
                ),
                "result": None,
            }
        )
        self.assertDictEqual(group.present("salt", delusers=True, members=True), ret)

        ret.update(
            {
                "comment": (
                    "Error. Same user(s) can not be added and deleted simultaneously"
                )
            }
        )
        self.assertDictEqual(group.present("salt", addusers=["a"], delusers=["a"]), ret)

        ret.update(
            {
                "comment": (
                    "The following group attributes are set"
                    " to be changed:\nkey0: value0\nkey1: value1\n"
                )
            }
        )

        mock = MagicMock(
            side_effect=[
                OrderedDict((("key0", "value0"), ("key1", "value1"))),
                False,
                False,
                False,
            ]
        )
        with patch.object(group, "_changes", mock):
            with patch.dict(group.__opts__, {"test": True}):
                self.assertDictEqual(group.present("salt"), ret)

                ret.update({"comment": "Group salt set to be added"})
                self.assertDictEqual(group.present("salt"), ret)

            with patch.dict(group.__opts__, {"test": False}):
                mock = MagicMock(return_value=[{"gid": 1, "name": "stack"}])
                with patch.dict(group.__salt__, {"group.getent": mock}):
                    ret.update(
                        {
                            "result": False,
                            "comment": (
                                "Group salt is not present but"
                                " gid 1 is already taken by group stack"
                            ),
                        }
                    )
                    self.assertDictEqual(group.present("salt", 1), ret)

                    mock = MagicMock(return_value=False)
                    with patch.dict(group.__salt__, {"group.add": mock}):
                        ret.update({"comment": "Failed to create new group salt"})
                        self.assertDictEqual(group.present("salt"), ret)

    def test_absent(self):
        """
        Test to ensure that the named group is absent
        """
        ret = {"name": "salt", "changes": {}, "result": True, "comment": {}}
        mock = MagicMock(side_effect=[True, True, True, False])
        with patch.dict(group.__salt__, {"group.info": mock}):
            with patch.dict(group.__opts__, {"test": True}):
                ret.update({"result": None, "comment": "Group salt is set for removal"})
                self.assertDictEqual(group.absent("salt"), ret)

            with patch.dict(group.__opts__, {"test": False}):
                mock = MagicMock(side_effect=[True, False])
                with patch.dict(group.__salt__, {"group.delete": mock}):
                    ret.update(
                        {
                            "result": True,
                            "changes": {"salt": ""},
                            "comment": "Removed group salt",
                        }
                    )
                    self.assertDictEqual(group.absent("salt"), ret)

                    ret.update(
                        {
                            "changes": {},
                            "result": False,
                            "comment": "Failed to remove group salt",
                        }
                    )
                    self.assertDictEqual(group.absent("salt"), ret)

            ret.update({"result": True, "comment": "Group not present"})
            self.assertDictEqual(group.absent("salt"), ret)



"""
Verify salt-ssh fails with a retcode > 0 when a pillar rendering fails.
"""

import pytest

from salt.defaults.exitcodes import EX_AGGREGATE

pytestmark = [
    pytest.mark.skip_on_windows(reason="salt-ssh not available on Windows"),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module", autouse=True)
def pillar_tree_render_fail(base_env_pillar_tree_root_dir):
    top_file = """
    base:
      'localhost':
        - fail_render
      '127.0.0.1':
        - fail_render
    """
    pillar_file = r"""
    not_defined: {{ abc }}
    """
    top_tempfile = pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    )
    pillar_tempfile = pytest.helpers.temp_file(
        "fail_render.sls", pillar_file, base_env_pillar_tree_root_dir
    )
    with top_tempfile, pillar_tempfile:
        yield


@pytest.mark.parametrize(
    "args",
    (
        ("state.sls", "basic"),
        ("state.highstate",),
        ("state.sls_id", "foo", "basic"),
        ("state.show_sls", "basic"),
        ("state.show_low_sls", "basic"),
        ("state.show_highstate",),
        ("state.show_lowstate",),
        ("state.top", "top.sls"),
    ),
)
def test_it(salt_ssh_cli, args):
    ret = salt_ssh_cli.run(*args)
    assert ret.returncode == EX_AGGREGATE
    assert isinstance(ret.data, list)
    assert ret.data
    assert isinstance(ret.data[0], str)
    assert ret.data[0] == "Pillar failed to render with the following messages:"
    assert ret.data[1].startswith("Rendering SLS 'fail_render' failed.")



import pytest

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


def test_listen_requisite(state, state_tree):
    """
    Tests a simple state using the listen requisite
    """
    sls_contents = """
    successful_changing_state:
      cmd.run:
        - name: echo "Successful Change"

    non_changing_state:
      test.succeed_without_changes

    test_listening_change_state:
      cmd.run:
        - name: echo "Listening State"
        - listen:
          - cmd: successful_changing_state

    test_listening_non_changing_state:
      cmd.run:
        - name: echo "Only run once"
        - listen:
          - test: non_changing_state

    # test that requisite resolution for listen uses ID declaration.
    # test_listening_resolution_one and test_listening_resolution_two
    # should both run.
    test_listening_resolution_one:
      cmd.run:
        - name: echo "Successful listen resolution"
        - listen:
          - cmd: successful_changing_state

    test_listening_resolution_two:
      cmd.run:
        - name: echo "Successful listen resolution"
        - listen:
          - cmd: successful_changing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        listener_state = (
            'cmd_|-listener_test_listening_change_state_|-echo "Listening'
            ' State"_|-mod_watch'
        )
        assert listener_state in ret

        absent_state = (
            'cmd_|-listener_test_listening_non_changing_state_|-echo "Only run'
            ' once"_|-mod_watch'
        )
        assert absent_state not in ret


def test_listen_in_requisite(state, state_tree):
    """
    Tests a simple state using the listen_in requisite
    """
    sls_contents = """
    successful_changing_state:
      cmd.run:
        - name: echo "Successful Change"
        - listen_in:
          - cmd: test_listening_change_state

    non_changing_state:
      test.succeed_without_changes:
        - listen_in:
          - cmd: test_listening_non_changing_state

    test_listening_change_state:
      cmd.run:
        - name: echo "Listening State"

    test_listening_non_changing_state:
      cmd.run:
        - name: echo "Only run once"

    # test that requisite resolution for listen_in uses ID declaration.
    # test_listen_in_resolution should run.
    test_listen_in_resolution:
      cmd.wait:
        - name: echo "Successful listen_in resolution"

    successful_changing_state_name_foo:
      test.succeed_with_changes:
        - name: foo
        - listen_in:
          - cmd: test_listen_in_resolution

    successful_non_changing_state_name_foo:
      test.succeed_without_changes:
        - name: foo
        - listen_in:
          - cmd: test_listen_in_resolution
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")

        listener_state = (
            'cmd_|-listener_test_listening_change_state_|-echo "Listening'
            ' State"_|-mod_watch'
        )
        assert listener_state in ret
        absent_state = (
            'cmd_|-listener_test_listening_non_changing_state_|-echo "Only run'
            ' once"_|-mod_watch'
        )
        assert absent_state not in ret


def test_listen_in_requisite_resolution(state, state_tree):
    """
    Verify listen_in requisite lookups use ID declaration to check for changes
    """
    sls_contents = """
    successful_changing_state:
      cmd.run:
        - name: echo "Successful Change"
        - listen_in:
          - cmd: test_listening_change_state

    non_changing_state:
      test.succeed_without_changes:
        - listen_in:
          - cmd: test_listening_non_changing_state

    test_listening_change_state:
      cmd.run:
        - name: echo "Listening State"

    test_listening_non_changing_state:
      cmd.run:
        - name: echo "Only run once"

    # test that requisite resolution for listen_in uses ID declaration.
    # test_listen_in_resolution should run.
    test_listen_in_resolution:
      cmd.wait:
        - name: echo "Successful listen_in resolution"

    successful_changing_state_name_foo:
      test.succeed_with_changes:
        - name: foo
        - listen_in:
          - cmd: test_listen_in_resolution

    successful_non_changing_state_name_foo:
      test.succeed_without_changes:
        - name: foo
        - listen_in:
          - cmd: test_listen_in_resolution
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        listener_state = (
            'cmd_|-listener_test_listen_in_resolution_|-echo "Successful listen_in'
            ' resolution"_|-mod_watch'
        )
        assert listener_state in ret


def test_listen_requisite_resolution(state, state_tree):
    """
    Verify listen requisite lookups use ID declaration to check for changes
    """
    sls_contents = """
    successful_changing_state:
      cmd.run:
        - name: echo "Successful Change"

    non_changing_state:
      test.succeed_without_changes

    test_listening_change_state:
      cmd.run:
        - name: echo "Listening State"
        - listen:
          - cmd: successful_changing_state

    test_listening_non_changing_state:
      cmd.run:
        - name: echo "Only run once"
        - listen:
          - test: non_changing_state

    # test that requisite resolution for listen uses ID declaration.
    # test_listening_resolution_one and test_listening_resolution_two
    # should both run.
    test_listening_resolution_one:
      cmd.run:
        - name: echo "Successful listen resolution"
        - listen:
          - cmd: successful_changing_state

    test_listening_resolution_two:
      cmd.run:
        - name: echo "Successful listen resolution"
        - listen:
          - cmd: successful_changing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        listener_state = (
            'cmd_|-listener_test_listening_resolution_one_|-echo "Successful listen'
            ' resolution"_|-mod_watch'
        )
        assert listener_state in ret


def test_listen_requisite_no_state_module(state, state_tree):
    """
    Tests a simple state using the listen requisite
    """
    sls_contents = """
    successful_changing_state:
      cmd.run:
        - name: echo "Successful Change"

    non_changing_state:
      test.succeed_without_changes

    test_listening_change_state:
      cmd.run:
        - name: echo "Listening State"
        - listen:
          - successful_changing_state

    test_listening_non_changing_state:
      cmd.run:
        - name: echo "Only run once"
        - listen:
          - non_changing_state

    # test that requisite resolution for listen uses ID declaration.
    # test_listening_resolution_one and test_listening_resolution_two
    # should both run.
    test_listening_resolution_one:
      cmd.run:
        - name: echo "Successful listen resolution"
        - listen:
          - successful_changing_state

    test_listening_resolution_two:
      cmd.run:
        - name: echo "Successful listen resolution"
        - listen:
          - successful_changing_state
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        listener_state = (
            'cmd_|-listener_test_listening_change_state_|-echo "Listening'
            ' State"_|-mod_watch'
        )
        assert listener_state in ret

        absent_state = (
            'cmd_|-listener_test_listening_non_changing_state_|-echo "Only run'
            ' once"_|-mod_watch'
        )
        assert absent_state not in ret


def test_listen_in_requisite_resolution_names(state, state_tree):
    """
    Verify listen_in requisite lookups use ID declaration to check for changes
    and resolves magic names state variable
    """
    sls_contents = """
    test:
      test.succeed_with_changes:
        - name: test
        - listen_in:
          - test: service

    service:
      test.succeed_without_changes:
        - names:
          - nginx
          - crond
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert "test_|-listener_service_|-nginx_|-mod_watch" in ret
        assert "test_|-listener_service_|-crond_|-mod_watch" in ret


def test_listen_requisite_resolution_names(state, state_tree):
    """
    Verify listen requisite lookups use ID declaration to check for changes
    and resolves magic names state variable
    """
    sls_contents = """
    test:
      test.succeed_with_changes:
        - name: test

    service:
      test.succeed_without_changes:
        - names:
          - nginx
          - crond
        - listen:
          - test: test
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert "test_|-listener_service_|-nginx_|-mod_watch" in ret
        assert "test_|-listener_service_|-crond_|-mod_watch" in ret


@pytest.mark.parametrize(
    "fun,onlyif,result,comment,assert_changes",
    (
        ("test.succeed_with_changes", [{}], True, "Success!", None),
        (
            "test.succeed_without_changes",
            [{"fun": "test.true"}],
            True,
            "Success!",
            False,
        ),
        (
            "test.fail_with_changes",
            [{"fun": "test.false"}],
            True,
            "onlyif condition is false",
            False,
        ),
        ("test.fail_with_changes", [{"fun": "test.true"}], False, "Failure!", True),
    ),
)
def test_onlyif_req(state, fun, onlyif, result, comment, assert_changes):
    ret = state.single(name="onlyif test", fun=fun, onlyif=onlyif)
    assert ret.result is result
    assert ret.comment == comment
    if assert_changes is True:
        assert ret.changes
    elif assert_changes is False:
        assert not ret.changes


def test_listen_requisite_not_exist(state, state_tree):
    """
    Tests a simple state using the listen requisite
    when the state id does not exist
    """
    sls_contents = """
    successful_changing_state:
      cmd.run:
        - name: echo "Successful Change"

    non_changing_state:
      test.succeed_without_changes

    test_listening_change_state:
      cmd.run:
        - name: echo "Listening State"
        - listen:
          - cmd: successful_changing_state

    test_listening_non_changing_state:
      cmd.run:
        - name: echo "Only run once"
        - listen:
          - test: non_changing_state_not_exist
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert (
            ret.raw[
                "Listen_Error_|-listen_non_changing_state_not_exist_|-listen_test_|-Listen_Error"
            ]["comment"]
            == "Referenced state test: non_changing_state_not_exist does not exist"
        )



import pytest

import salt.utils.win_lgpo_netsh as win_lgpo_netsh
from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
]


@pytest.mark.parametrize("store", ["local", "lgpo"])
def test_get_settings_firewallpolicy(store):
    ret = win_lgpo_netsh.get_settings(
        profile="domain", section="firewallpolicy", store=store
    )
    assert "Inbound" in ret
    assert "Outbound" in ret


@pytest.mark.parametrize("store", ["local", "lgpo"])
def test_get_settings_logging(store):
    ret = win_lgpo_netsh.get_settings(profile="domain", section="logging", store=store)
    assert "FileName" in ret
    assert "LogAllowedConnections" in ret
    assert "LogDroppedConnections" in ret
    assert "MaxFileSize" in ret


@pytest.mark.parametrize("store", ["local", "lgpo"])
def test_get_settings_settings(store):
    ret = win_lgpo_netsh.get_settings(profile="domain", section="settings", store=store)
    assert "InboundUserNotification" in ret
    assert "LocalConSecRules" in ret
    assert "LocalFirewallRules" in ret
    assert "UnicastResponseToMulticast" in ret


@pytest.mark.parametrize("store", ["local", "lgpo"])
def test_get_settings_state(store):
    ret = win_lgpo_netsh.get_settings(profile="domain", section="state", store=store)
    assert "State" in ret


@pytest.mark.parametrize("store", ["local", "lgpo"])
def test_get_all_settings(store):
    ret = win_lgpo_netsh.get_all_settings(profile="domain", store=store)
    assert "Inbound" in ret
    assert "Outbound" in ret
    assert "FileName" in ret
    assert "LogAllowedConnections" in ret
    assert "LogDroppedConnections" in ret
    assert "MaxFileSize" in ret
    assert "InboundUserNotification" in ret
    assert "LocalConSecRules" in ret
    assert "LocalFirewallRules" in ret
    assert "UnicastResponseToMulticast" in ret
    assert "State" in ret


@pytest.mark.parametrize("store", ["local", "lgpo"])
def test_get_all_profiles(store):
    ret = win_lgpo_netsh.get_all_profiles(store=store)
    assert "Domain Profile" in ret
    assert "Private Profile" in ret
    assert "Public Profile" in ret


@pytest.mark.destructive_test
@pytest.mark.parametrize("store", ["local", "lgpo"])
@pytest.mark.parametrize(
    "inbound", ["allowinbound", "blockinbound", "blockinboundalways", "notconfigured"]
)
def test_set_firewall_settings_inbound(store, inbound):
    if inbound == "notconfigured" and store == "local":
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_firewall_settings,
            profile="domain",
            inbound=inbound,
            store=store,
        )
    else:
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store=store
        )["Inbound"]
        try:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", inbound=inbound, store=store
            )
            assert ret is True
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="firewallpolicy", store=store
            )["Inbound"]
            assert new.lower() == inbound
        finally:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", inbound=current, store=store
            )
            assert ret is True


@pytest.mark.destructive_test
@pytest.mark.parametrize("store", ["local", "lgpo"])
@pytest.mark.parametrize(
    "outbound", ["allowoutbound", "blockoutbound", "notconfigured"]
)
def test_set_firewall_settings_outbound(store, outbound):
    if outbound == "notconfigured" and store == "local":
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_firewall_settings,
            profile="domain",
            inbound=outbound,
            store=store,
        )
    else:
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="firewallpolicy", store=store
        )["Outbound"]
        try:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", outbound=outbound, store=store
            )
            assert ret is True
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="firewallpolicy", store=store
            )["Outbound"]
            assert new.lower() == outbound
        finally:
            ret = win_lgpo_netsh.set_firewall_settings(
                profile="domain", outbound=current, store=store
            )
            assert ret is True


@pytest.mark.destructive_test
@pytest.mark.parametrize("store", ["local", "lgpo"])
@pytest.mark.parametrize("setting", ["allowedconnections", "droppedconnections"])
@pytest.mark.parametrize("value", ["enable", "disable", "notconfigured"])
def test_set_firewall_logging_connections(store, setting, value):
    if value == "notconfigured" and store == "local":
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_logging_settings,
            profile="domain",
            setting=setting,
            value=value,
            store=store,
        )
    else:
        setting_map = {
            "allowedconnections": "LogAllowedConnections",
            "droppedconnections": "LogDroppedConnections",
        }
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store=store
        )[setting_map[setting]]
        try:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting=setting,
                value=value,
                store=store,
            )
            assert ret is True
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="logging", store=store
            )[setting_map[setting]]
            assert new.lower() == value
        finally:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain",
                setting=setting,
                value=current,
                store=store,
            )
            assert ret is True


@pytest.mark.destructive_test
@pytest.mark.parametrize("store", ["local", "lgpo"])
@pytest.mark.parametrize("value", ["C:\\Temp\\test.log", "notconfigured"])
def test_set_firewall_logging_filename(store, value):
    current = win_lgpo_netsh.get_settings(
        profile="domain", section="logging", store=store
    )["FileName"]
    try:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain",
            setting="filename",
            value=value,
            store=store,
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store=store
        )["FileName"]
        assert new.lower() == value.lower()
    finally:
        ret = win_lgpo_netsh.set_logging_settings(
            profile="domain", setting="filename", value=current, store=store
        )
        assert ret is True


@pytest.mark.destructive_test
@pytest.mark.parametrize("store", ["local", "lgpo"])
@pytest.mark.parametrize("value", ["16384", "notconfigured"])
def test_set_firewall_logging_maxfilesize(store, value):
    if value == "notconfigured":
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_logging_settings,
            profile="domain",
            setting="maxfilesize",
            value=value,
            store=store,
        )
    else:
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="logging", store=store
        )["MaxFileSize"]
        try:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain", setting="maxfilesize", value=value, store=store
            )
            assert ret is True
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="logging", store=store
            )["MaxFileSize"]
            assert new == int(value)
        finally:
            ret = win_lgpo_netsh.set_logging_settings(
                profile="domain", setting="maxfilesize", value=current, store=store
            )
            assert ret is True


@pytest.mark.destructive_test
@pytest.mark.parametrize("store", ["local", "lgpo"])
@pytest.mark.parametrize(
    "setting",
    ["localconsecrules", "inboundusernotification", "unicastresponsetomulticast"],
)
@pytest.mark.parametrize("value", ["enable", "disable", "notconfigured"])
def test_set_firewall_settings(store, setting, value):
    setting_map = {
        "localconsecrules": "LocalConSecRules",
        "inboundusernotification": "InboundUserNotification",
        "unicastresponsetomulticast": "UnicastResponseToMulticast",
    }
    if value == "notconfigured" and store == "local":
        pytest.raises(
            CommandExecutionError,
            win_lgpo_netsh.set_settings,
            profile="domain",
            setting=setting,
            value=value,
            store=store,
        )
    else:
        current = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store=store
        )[setting_map[setting]]
        try:
            ret = win_lgpo_netsh.set_settings(
                profile="domain",
                setting=setting,
                value=value,
                store=store,
            )
            assert ret is True
            new = win_lgpo_netsh.get_settings(
                profile="domain", section="settings", store=store
            )[setting_map[setting]]
            assert new.lower() == value
        finally:
            if current != "notconfigured":
                ret = win_lgpo_netsh.set_settings(
                    profile="domain",
                    setting=setting,
                    value=current,
                    store=store,
                )
            assert ret is True


@pytest.mark.destructive_test
@pytest.mark.parametrize("store", ["local", "lgpo"])
@pytest.mark.parametrize("state", ["on", "off", "notconfigured"])
def test_set_firewall_state(store, state):
    current_state = win_lgpo_netsh.get_settings(
        profile="domain", section="state", store=store
    )["State"]
    try:
        ret = win_lgpo_netsh.set_state(profile="domain", state=state, store=store)
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store=store
        )["State"]
        assert new.lower() == state.lower()
    finally:
        win_lgpo_netsh.set_state(profile="domain", state=current_state, store=store)


@pytest.mark.destructive_test
@pytest.mark.parametrize("store", ["local", "lgpo"])
@pytest.mark.parametrize("allow_inbound", ["enable", "disable"])
@pytest.mark.parametrize("state", ["on", "off", "notconfigured"])
def test_set_firewall_state_allow_inbound(store, allow_inbound, state):
    current_state = win_lgpo_netsh.get_settings(
        profile="domain", section="state", store=store
    )["State"]
    current_local_fw_rules = win_lgpo_netsh.get_settings(
        profile="domain", section="settings", store="lgpo"
    )["LocalFirewallRules"]
    try:
        ret = win_lgpo_netsh.set_settings(
            profile="domain",
            setting="localfirewallrules",
            value=allow_inbound,
            store=store,
        )
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="settings", store=store
        )["LocalFirewallRules"]
        assert new.lower() == allow_inbound.lower()
        ret = win_lgpo_netsh.set_state(profile="domain", state=state, store=store)
        assert ret is True
        new = win_lgpo_netsh.get_settings(
            profile="domain", section="state", store=store
        )["State"]
        assert new.lower() == state.lower()
    finally:
        if current_local_fw_rules.lower() != "notconfigured":
            win_lgpo_netsh.set_settings(
                profile="domain",
                setting="localfirewallrules",
                value=current_local_fw_rules,
                store=store,
            )
        win_lgpo_netsh.set_state(profile="domain", state=current_state, store=store)



"""
Send events from webhook api
"""

import tornado.httpserver
import tornado.ioloop
import tornado.web

import salt.utils.event


def start(address=None, port=5000, ssl_crt=None, ssl_key=None):
    """
    Api to listen for webhooks to send to the reactor.

    Implement the webhook behavior in an engine.
    :py:class:`rest_cherrypy Webhook docs <salt.netapi.rest_cherrypy.app.Webhook>`

    Unlike the rest_cherrypy Webhook, this is only an unauthenticated webhook
    endpoint.  If an authenticated webhook endpoint is needed, use the salt-api
    webhook which runs on the master and authenticates through eauth.

    .. note: This is really meant to be used on the minion, because salt-api
             needs to be run on the master for use with eauth.

    .. warning:: Unauthenticated endpoint

        This engine sends webhook calls to the event stream.  If the engine is
        running on a minion with `file_client: local` the event is sent to the
        minion event stream.  Otherwise it is sent to the master event stream.

    Example Config

    .. code-block:: yaml

        engines:
          - webhook: {}

    .. code-block:: yaml

        engines:
          - webhook:
              port: 8000
              address: 10.128.1.145
              ssl_crt: /etc/pki/tls/certs/localhost.crt
              ssl_key: /etc/pki/tls/certs/localhost.key

    .. note: For making an unsigned key, use the following command
             `salt-call --local tls.create_self_signed_cert`
    """
    if __opts__.get("__role") == "master":
        fire_master = salt.utils.event.get_master_event(
            __opts__, __opts__["sock_dir"]
        ).fire_event
    else:
        fire_master = None

    def fire(tag, msg):
        """
        How to fire the event
        """
        if fire_master:
            fire_master(msg, tag)
        else:
            __salt__["event.send"](tag, msg)

    class WebHook(tornado.web.RequestHandler):  # pylint: disable=abstract-method
        def post(self, tag):  # pylint: disable=arguments-differ
            body = self.request.body
            headers = self.request.headers
            payload = {
                "headers": headers if isinstance(headers, dict) else dict(headers),
                "body": body,
            }
            fire("salt/engines/hook/" + tag, payload)

    application = tornado.web.Application([(r"/(.*)", WebHook)])
    ssl_options = None
    if all([ssl_crt, ssl_key]):
        ssl_options = {"certfile": ssl_crt, "keyfile": ssl_key}
    io_loop = tornado.ioloop.IOLoop()
    http_server = tornado.httpserver.HTTPServer(application, ssl_options=ssl_options)
    http_server.listen(port, address=address)
    io_loop.start()



import pytest

import salt.modules.http as http


@pytest.fixture
def configure_loader_modules():
    return {http: {}}


def test_query_error_tornado(webserver):
    """
    Ensure we return expected data when website does not return
    a 200 with tornado
    """
    url = webserver.url("doesnotexist")
    ret = http.query(url, backend="tornado")
    assert (
        ret["body"]
        == "<html><title>404: Not Found</title><body>404: Not Found</body></html>"
    )
    assert ret["error"] == "HTTP 404: Not Found"
    assert ret["status"] == 404


@pytest.mark.parametrize("backend", ["requests", "urllib2", "tornado"])
def test_query_success(webserver, backend):
    """
    Ensure we return a success when querying
    a website
    """
    url = webserver.url("this.txt")
    ret = http.query(url, backend=backend)
    assert ret == {"body": "test"}
